{+-----------------------------------------------------------------------------+
 | Class:       TmwCustomEdit
 | Created:     11.98
 | Last change: 17.12.98
 | Author:      Martin Waldenburg
 | Description: study on how to create a custom edit control without using
 |              a Windows edit control.
 | Version:     0.7 beta (version history follows licence agreement)
 | Copyright (c) 1998 Martin Waldenburg
 | All rights reserved.
 |
 | Thanks to : Woo Young Bum, Angus Johnson, Michael Trier, James Jacobson,
 |             Thomas Kurz, Primoz Gabrijelcic, Michael Beck, Andy Jeffries,
 |             Edward Kreis, Brad Stowers
 |
 | LICENCE CONDITIONS
 |
 | USE OF THE ENCLOSED SOFTWARE
 | INDICATES YOUR ASSENT TO THE
 | FOLLOWING LICENCE CONDITIONS.
 |
 |
 |
 | These Licence Conditions are exlusively
 | governed by the Law and Rules of the
 | Federal Republic of Germany.
 |
 | Redistribution and use in source and binary form, with or without
 | modification, are permitted provided that the following conditions
 | are met:
 |
 | 1. Redistributions of source code must retain the above copyright
 |    notice, this list of conditions and the following disclaimer.
 |    If the source is modified, the complete original and unmodified
 |    source code has to distributed with the modified version.
 |
 | 2. Redistributions in binary form must reproduce the above
 |    copyright notice, these licence conditions and the disclaimer
 |    found at the end of this licence agreement in the documentation
 |    and/or other materials provided with the distribution.
 |
 | 3. Software using this code must contain a visible line of credit.
 |
 | 4. If my code is used in a "for profit" product, you have to donate
 |    to a registered charity in an amount that you feel is fair.
 |    You may use it in as many of your products as you like.
 |    Proof of this donation must be provided to the author of
 |    this software.
 |
 | 5. If you for some reasons don't want to give public credit to the
 |    author, you have to donate three times the price of your software
 |    product, or any other product including this component in any way,
 |    but no more than $500 US and not less than $200 US, or the
 |    equivalent thereof in other currency, to a registered charity.
 |    You have to do this for every of your products, which uses this
 |    code separately.
 |    Proof of this donations must be provided to the author of
 |    this software.
 |
 |
 | DISCLAIMER:
 |
 | THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS IS'.
 |
 | ALL EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 | PARTICULAR PURPOSE ARE DISCLAIMED.
 |
 | IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 | INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 | (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 | OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 | INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 | WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 | NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 | THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 |
 |  Martin.Waldenburg@T-Online.de
 |
 | Known problems:
 |   - dragging cannot be canceled with the chord clicking - can't fix!
 |
 | Version history:
 |   up to 0.53: source maintained by the author, version history unknown
 |   0.6: 9.12.1998
 |     - Primoz Gabrijelcic (primoz.gabrijelcic@altavista.net) became a source
 |       maintainer for mwCustomEdit and mwPasSyn
 |     - main .pas files renamed to mwCustomEdit and mwPasSyn to enhance
 |       clarity
 |     - Thomas Kurz: added ability to extend selection by pressing Shift before
 |       clicking
 |     - James Jacobson: smoothing the caret movement; new property and class
 |       to allow changing the Selected text colour; fix to prevent putting name
 |       in control when loading in IDE and Text is empty
 |     - Primoz Gabrijelcic: removed one warning in TmwCustomEdit.CreateParams
 |     - mwCustomEdit got a web page
 |       (http://www.eccentrica.org/gabr/mw/mwedit.htm, mirror on
 |       http://members.xoom.com/primozg/mw/mwedit.htm) and a mailing list
 |       (mwedit@listbot.com, subscription form available on mwCustomEdit
 |       web page)
 |   0.61: 10.12.1998
 |     - Michael Beck: created a help file, generated with Time2Help
 |       (http://www.time2help.com)
 |   0.62: 10.12.1998
 |     - mwPasSyn updated (current version: 0.54)
 |     - Andy Jeffries: donated file mirror space on Kwikrite
 |       (http://www.kwikrite.clara.net/mwEdit/mwedit.zip)
 |     - Thomaz Kurz: added Undo capability (property CanUndu and procedure
 |       Undu)
 |   0.7:
 |     - mwPasSyn updated (current version: 0.6)
 |     - James Jacobson: bookmarks (10 bookmarks, Ctrl+Shift+<number>
 |       sets/clears, Ctrl+<number> jumps, images hardcoded, exposed through
 |       property TmwBookMarkOpt); some fixes dealing with the caret,
 |       unselecting text, scrollbars, backspacing and deleting when the line
 |       length is zero; a few Set methods and Default values (fixed invalid
 |       behaviour when GutterWidth was changed); OnPaint event
 |     - Edward Kreis: added ability to move or copy selected text by drag&drop
 |     - Primoz Gabrijelcic: hardwired Undo to the Alt+Backspace; added some
 |       event debugging code; changed drag&drop code to allow single click in
 |       the middle of selection to clear selection (fixed strayed MouseUp
 |       procession too); added overriden Notification to beter handle removal
 |       of associated highlighter; added Set method for Highlighter property
 |       to beter handle removal of associated highlighter; general
 |       reformatting & cleanup; created common registration unit mwEditReg
 |       (current version: 0.11); fixed drag&drop threshold to work in other
 |       compilers beside D4; improved Delphi demo; fixed Left key behaviour:
 |       when ScrollPastEOL = false, Left moves caret from beginning of the line
 |       to the end of previous line; large rewrite of ScrollPastEOL; fix for
 |       invalid drag&drop behaviour when more than one line was moved 
 |     - Michael Trier: hacked mwCustomEdit (a lot) to compile with BCB 3;
 |       donated file mirror space on Dcj FTP Site and Dcj FTP Mirror Site
 |       (http://www.dcjournal.com/ftp/tidbits/mwedit.zip and
 |       ftp://ftp.safetystorage.com/pub/safetystor/journal/tidbits/mwedit.zip);
 |       added C++ syntax highlighter TDcjCppSyn (DcjCppSyn.pas; current
 |       version: 0.61); renamed values in TChangeReason; added BCB 3 demo;
 |       added BCB 1 demo
 |     - Thomas Kurz: bugfixes for Undo, MaxUndo property, OnChange
 |     - Brad Stowers: fixed Ctrl+Shift+End to select last line too; limited
 |       Font property dialog (IDE) to just fixed fonts; added BorderStyle
 |       property; published inherited properties ParentCtl3D, Anchors (D4), and
 |       Contstraints (D4); added HideSelection property; better cursor position
 |       for Ctrl-End; fix: Home, End, PgUp, PgDn would not remove selection if
 |       pressed without Shift key; added Ctrl-Up and Ctrl-Down keyboard
 |       commands:  Scrolls editor up or down one line without moving caret
 |       position (unless the caret would not be visible) - same as Delphi IDE;
 |       InsertMode published property for fInserting variable; overwrite mode
 |       caret; added property ScrollPastEOL to (dis)allow scrolling past EOL
 |     - Martin Waldenburg: fixed buggy TmwCustomEdit.ListPutted
 |
 +----------------------------------------------------------------------------+}

{$I MWEDIT.INC}  //mt 12/16/1998 added
//mt 12/16/1998 removed {$ObjExportAll On} since it's defined in the include

unit mwCustomEdit;

interface

uses
// Nasty BCB linker error when doing it this way:                //mt 12/10/1998
//   Windows, WinProcs, Messages, Classes, Controls, Printers,
//   Forms, Graphics, SysUtils, StdCtrls, Clipbrd, ExtCtrls, mwPasSyn;
  Windows, Messages, SysUtils, Classes, Controls, Graphics,
  ExtCtrls, Forms, StdCtrls, Clipbrd, mwHighlighter;


const
   DIGIT            = ['0'..'9'];
// ALPHA            = ['A'..'Z', 'a'..'z'];                      //mt 12/10/1998
// break these up because we exceed the 4 byte limit when combined.
   ALPHA_UC         = ['A'..'Z'];
   ALPHA_LC         = ['a'..'z'];

type
  TIndexEvent = procedure(Index: Integer) of object;
  TPaintEvent = procedure(Sender:TObject; ACanvas:TCanvas)      //jdj 12/12/1998
                  of object;

  {begin}                                                    //tskurz 12/10/1998
  // added mw as a prefix otherwise crNone conflicts with the    //mt 12/16/1998
  // crNone cursor constant.
  TChangeReason = (mwcrInsert, mwcrDragDropInsert, mwcrDelete,
                   mwcrDragDropDelete, mwcrLineBreak, mwcrNone);

  TChangePtr = ^TChange;
  TChange = record
              ChangeStr: PChar;
              ChangeReason: TChangeReason;
              ChangeStartPos,
              ChangeEndPos: TPoint;
            end;
            
  //had to move it up here or the header creator on BCB bombs    //mt 12/16/1998
  {+----------------------------------------------+
   |               -Bookmarks-
   |There are 10 available numbered 0 to 9
   |The X and Y values are 1 based as are the
   |  CaretX and CaretY properties
   |Shift+Ctrl+TheBookMarkNumber will Set/Clear
   |Ctrl+TheBookMarkNumber will Goto
   |The bitmaps are in mwCustomEdit.res and
   |  are named MARK_ + TheBookMarkNumber
  +----------------------------------------------+}
  TBookMarkRec = record                                         //jdj 12/10/1998
    IsSet: Boolean;
    X: Integer;
    Y: Integer;
    Glyph: TBitmap;
  end;

  TmwCustomEdit = class;                                     //tskurz 12/15/1998

  TUndoList = class
    private
      FList: TList;
      FCanUndo: Integer;
      FMaxUndo: Integer;
      FOwner: TmwCustomEdit;                                 //tskurz 12/15/1998
      function GetCanUndo: Integer;
      procedure SetMaxUndo(const Value: Integer);
    protected
      procedure RemoveChange(index: Integer);
    public
      constructor Create(AOwner: TmwCustomEdit);             //tskurz 12/15/1998
      destructor Destroy; override;
      procedure AddChange(ChangeReason: TChangeReason; ChangeStartPos,
                 ChangeEndPos: TPoint; ChangeStr: PChar);
      function GetChange(var ChangeStartPos, ChangeEndPos: TPoint;
                 var ChangeStr: PChar): TChangeReason;
      function GetChangeReason: TChangeReason;               //tskurz 12/16/1998
    published
      property CanUndo : Integer read GetCanUndo;
      property MaxUndo : Integer read FMaxUndo write SetMaxUndo;
  end;
  {end}                                                      //tskurz 12/10/1998

  TmwEditList = class(TStringList)
  private
    FOnAdded: TNotifyEvent;
    FOnDeleted: TIndexEvent;
    FOnInserted: TIndexEvent;
    FOnLoaded: TNotifyEvent;
    fOnPutted: TIndexEvent;
  protected
    procedure Put(Index: Integer; const S: string); override;
  public
    function Add(const S: string): Integer; override;
    procedure Delete(Index: Integer); override;
    procedure Insert(Index: Integer; const S: string); override;
    procedure SetTextStr(const Value: string); override;
    property OnAdded: TNotifyEvent read FOnAdded write FOnAdded;
    property OnDeleted: TIndexEvent read FOnDeleted write FOnDeleted;
    property OnInserted: TIndexEvent read FOnInserted write FOnInserted;
    property OnLoaded: TNotifyEvent read FOnLoaded write FOnLoaded;
    property OnPutted: TIndexEvent read FOnPutted write FOnPutted;
  published
  end;

  TmwSelectedColor = class(TPersistent)
  private
    FBG: TColor;
    FFG: TColor;
    FOwner: TmwCustomEdit;
    procedure SetBG(Value: TColor);
    procedure SetFG(Value: TColor);
  public
    constructor Create(AOwner: TmwCustomEdit);
  published
    property BG: TColor read FBG write SetBG default clHighLight;
    property FG: TColor read FFG write SetFG default clHighLightText;
  end; 

  TmwBookMarkOpt = class(TPersistent)                           //jdj 12/12/1998
  private
    fEnableKeys: Boolean;
    fGlyphsVisible: Boolean;
    fLeftMargin: Integer;
    FOwner: TmwCustomEdit;
    procedure SetGlyphsVisible(Value: Boolean);
    procedure SetLeftMargin(Value: Integer);
  public
    constructor Create(AOwner: TmwCustomEdit);
  published
    property EnableKeys: Boolean read fEnableKeys write fEnableKeys default True;
    property GlyphsVisible: Boolean read fGlyphsVisible write SetGlyphsVisible default True;
    property LeftMargin: Integer read fLeftMargin write SetLeftMargin default 2;
  end;

  TmwCustomEdit = class(TCustomControl)
  private
    fBlockBegin: TPoint;
    fBlockEnd: TPoint;
    fBottomLine: Integer;
    fCaretX: Integer;
    fCaretY: Integer;
    fCaretYPix: Integer;
    fCaretVisible: Boolean;
    fCharsInWindow: Integer;
    fCharWidth: Integer;
    fDblClicked: Boolean;
    fEditRect: TRect;
    fEditTimer: TTimer;
    fElapsed: Boolean;
    fFontDummy: TFont;
    fGutterColor: TColor;
    fGutterRect: TRect;
    fGutterWidth: Integer;
    fInEditRect: Boolean;
    fInserting: Boolean;
    fLineRect: TRect;
    fLines: TStrings;
    fLinesInWindow: Integer;
    fLeftChar: Integer;
    fMaxLeftChar: Integer;
    fPaintLock: Integer;
    fReadOnly: Boolean;
    fRightChar: Integer;
    fRightEdge: Integer;
    fRightEdgeOffset: Integer;
    FScrollBars: TScrollStyle;
    fTextHeight: Integer;
    fTextOffset: Integer;
    fTopLine: Integer;
    fHighLighter: TmwCustomHighLighter;
    fSelectedColor: TmwSelectedColor;                           //jdj 12/09/1998
    fUndoList : TUndoList;                                   //tskurz 12/10/1998
    fBookMarks: array[0..9] of TBookMarkRec;                    //jdj 12/10/1998
    fDragBlockBegin: TPoint;                                    //ked 12/10/1998
    fDragBlockEnd: TPoint;                                      //ked 12/10/1998
    fMouseDownX: integer;                                       //ked 12/10/1998
    fMouseDownY: integer;                                       //ked 12/10/1998
    fWaitForDragging: boolean;                                   //gp 12/12/1998
    fBookMarkOpt: TmwBookMarkOpt;                               //jdj 12/12/1998
    fOnPaint: TPaintEvent;                                      //jdj 12/12/1998
    fOnChange: TNotifyEvent;                                 //tskurz 12/15/1998
    fBorderStyle: TBorderStyle;                                 //bds 12/16/1998 
    fHideSelection: boolean;                                    //bds 12/16/1998 
    fScrollPastEOL: boolean;                                    //bds 12/16/1998
    function CaretXPix: Integer;
    function CaretYPix: Integer;
    procedure ComputeCaret(X, Y: Integer);
    function GetBlockBegin: TPoint;
    function GetBlockEnd: TPoint;
    function GetCaretX: Integer;
    function GetCaretY: Integer;
    function GetFont: TFont;
    function GetLeftChar: Integer;
    function GetLineCount: Integer;
    function GetLineText: String;
    function GetSelAvail: Boolean;
    function GetText: String;
    function GetTopLine: Integer;
    procedure FontChanged(Sender: TObject);
    function LeftSpaces(Line: String): Integer;
    procedure LinesChanged(Sender: TObject);
    procedure PaintControl;
    procedure SetBlockBegin(Value: TPoint);
    procedure SetBlockEnd(Value: TPoint);
    procedure SetWordBlock(Value: TPoint);
    procedure SetCaretX(Value: Integer);
    procedure SetCaretY(Value: Integer);
    procedure SetGutterWidth(Value: Integer);
    procedure SetFont(const Value: TFont);
    procedure SetLeftChar(Value: Integer);
    procedure SetLines(Value: TStrings);
    procedure SetLineText(Value: String);
    procedure SetScrollBars(const Value: TScrollStyle);
    procedure SetText(const Value: String);
    procedure SetTopLine(Value: Integer);
    procedure TimerFired(Sender: TObject);
    procedure UpdateCaret;
    procedure UpdateScrollBars;
    function GetSelText: String;
    procedure SetSelText(const Value: String);
    procedure ScanFrom(Index: Integer);
    procedure PaintHighLighted(TextBM: TBitMap);
    function GetCanUndo : Boolean;                           //tskurz 12/10/1998
    procedure SetGutterColor(Value: TColor);                    //jdj 12/12/1998
    procedure SetRightEdge(Value: Integer);                     //jdj 12/12/1998
    procedure SetMaxUndo(const Value: Integer);
    function GetMaxUndo: Integer;                            //tskurz 12/14/1998
    procedure SetHighlighter(const Value: TmwCustomHighLighter); //gp 12/16/1998
    procedure SetBorderStyle(Value: TBorderStyle);              //bds 12/16/1998
    procedure SetHideSelection(const Value: boolean);           //bds 12/16/1998
    procedure SetInsertMode(const Value: boolean);              //bds 12/16/1998
    procedure SetScrollPastEOL(const Value: boolean);           //bds 12/16/1998
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure DecPaintLock;
    procedure HideCaret;
    procedure IncPaintLock;
    procedure ListAdded(Sender: TObject);
    procedure ListDeleted(Index: Integer);
    procedure ListInserted(Index: Integer);
    procedure ListLoaded(Sender: TObject);
    procedure ListPutted(Index: Integer);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure DblClick; override;
    procedure KeyPress(var Key: Char); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure PaintGutter(TextBM: TBitMap); virtual;
    procedure PaintText(TextBM: TBitMap); virtual;
    procedure Paint; override;
    procedure ProcessVK_BACK(Shift: TShiftState); virtual;
    procedure ProcessVK_DELETE(Shift: TShiftState); virtual;
    procedure ProcessVK_DOWN(Shift: TShiftState); virtual;
    procedure ProcessVK_END(Shift: TShiftState); virtual;
    procedure ProcessVK_HOME(Shift: TShiftState); virtual;
    procedure ProcessVK_INSERT(Shift: TShiftState); virtual;
    procedure ProcessVK_LEFT(Shift: TShiftState); virtual;
    procedure ProcessVK_NEXT(Shift: TShiftState); virtual;
    procedure ProcessVK_PRIOR(Shift: TShiftState); virtual;
    procedure ProcessVK_RETURN(Shift: TShiftState); virtual;
    procedure ProcessVK_RIGHT(Shift: TShiftState); virtual;
    procedure ProcessVK_UP(Shift: TShiftState); virtual;
    procedure SetName(const Value: TComponentName); override;
    procedure ShowCaret;
    procedure Loaded; override;                                 //jdj 12/12/1998
    function CoorToIndex(CPos: TPoint): integer;                //ked 12/10/1998
    function IndexToCoor(ind: integer): TPoint;                 //ked 12/10/1998
    procedure DragOver(Source: TObject; X, Y: Integer;          //ked 12/10/1998
                       State: TDragState; var Accept: Boolean); override;
    procedure InitializeCaret;                                  //bds 12/16/1998                       
    procedure WMEraseBkgnd(var Message: TMessage); message WM_ERASEBKGND;
    procedure WMGetDlgCode(var Msg: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMHScroll(var Message: TWMScroll); message WM_HSCROLL;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMNCHitTest(var Message: TWMNCHitTest); message WM_NCHITTEST;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMVScroll(var Message: TWMScroll); message WM_VSCROLL;
    property PaintLock: Integer read fPaintLock;
    property SelAvail: Boolean read GetSelAvail;
  public
    Constructor Create(AOwner: TComponent); override;
    Destructor Destroy; override;
    procedure WndProc(var msg: TMessage); override;
    procedure DragDrop(Source: TObject; X, Y: Integer);         //ked 12/10/1998
                override;
    procedure CopyToClipboard;
    procedure CutToClipboard;
    procedure PasteFromClipboard;
    procedure SelectAll;
    procedure Undo;                                          //tskurz 12/10/1998
    procedure ClearBookMark(BookMark: Integer);                 //jdj 12/10/1998
    procedure GotoBookMark(BookMark: Integer);                  //jdj 12/10/1998
    procedure SetBookMark(BookMark: Integer; X: Integer; Y: Integer);
                                                                //jdj 12/10/1998
    procedure Notification(AComponent: TComponent;               //gp 12/16/1998
                Operation: TOperation); override;
    property Elapsed: Boolean read fElapsed;
    property BlockBegin: TPoint read GetBlockBegin write SetBlockBegin;
    property BlockEnd: TPoint read GetBlockEnd write SetBlockEnd;
    property CaretX: Integer read GetCaretX write SetCaretX;
    property CaretY: Integer read GetCaretY write SetCaretY;
    property CharsInWindow: Integer read fCharsInWindow;
    property LinesInWindow: Integer read fLinesInWindow;
    property SelText: String read GetSelText write SetSelText;
    property Text: String read GetText write SetText;
  published
    property Align;
    {begin}                                                     //bds 12/16/1998
    {$IFDEF MWE_COMPILER_4_UP}
    property Anchors;
    property Constraints;
    {$ENDIF}
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle
               default bsSingle;
    {end}                                                       //bds 12/16/1998
    property BookMarkOptions: TmwBookMarkOpt                    //jdj 12/12/1998
               read fBookMarkOpt write fBookMarkOpt;
    property CanUndo: Boolean read GetCanUndo;
    property Color;
    property Ctl3D;
    property Enabled;
    property Font: TFont read GetFont write SetFont;
    property GutterColor: TColor                      //jdj 12/12/1998 added Set
               read fGutterColor write SetGutterColor;
    property GutterWidth: Integer read fGutterWidth write SetGutterWidth;
    property Height;
    property HideSelection: boolean read fHideSelection         //bds 12/16/1998
               write SetHideSelection default false;
    property HighLighter: TmwCustomHighLighter       //gp 12/16/1998 - added Set
               read fHighLighter write SetHighlighter;
    property InsertMode: boolean read fInserting                //bds 12/16/1998
               write SetInsertMode default true;
    property LeftChar: Integer read GetLeftChar write SetLeftChar;
    property Lines: TStrings read fLines write SetLines;
    property LineCount: Integer read GetLineCount;
    property LineText: String read GetLineText write SetLineText;
    property MaxUndo : Integer read GetMaxUndo write SetMaxUndo;
                                                             //tskurz 12/14/1998
    property Name;
    property ParentColor;
    property ParentCtl3D;                                       //bds 12/16/1998
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly: Boolean read fReadOnly write fReadOnly;
    property RightEdge: Integer               //jdj 12/12/1998 added Set-Default
               read fRightEdge write SetRightEdge default 80;
    property ShowHint;
    property ScrollBars: TScrollStyle
               read FScrollBars write SetScrollBars default ssBoth;
    property ScrollPastEOL: boolean read FScrollPastEOL
               write SetScrollPastEOL default true;             //bds 12/16/1998
    property SelectedColor: TmwSelectedColor                    //jdj 12/09/1998
               read FSelectedColor write FSelectedColor;
    property TabOrder;
    property TabStop default True;                //jdj 12/12/1998 Added Default
    property Tag;
    property TopLine: Integer read GetTopLine write SetTopLine;
    property Visible;
    property Width;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
                                                             //tskurz 12/15/1998
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnPaint: TPaintEvent read fOnPaint write fOnPaint; //jdj 12/12/1998
    property OnStartDrag;
  end;

procedure Register;

implementation

{$R mwCustomEdit.RES}                                           //jdj 12/10/1998

procedure Register;
begin
  RegisterComponents('mw', [TmwCustomEdit]);
end;

function max(x, y: integer): integer;
begin
  if x > y then
    result := x else
    result := y;
end;

function min(x, y: integer): integer;
begin
  if x < y then
    result := x else
    result := y;
end;

function maxPoint(P1, P2: TPoint): TPoint;                   //tskurz 12/14/1998
begin
  if P1.y > P2.y then
  begin
    result := P1;
    exit;
  end else
    if P1.y < P2.y then
    begin
      result := P2;
      exit;
    end;
  if P1.x > P2.x then
    result := P1 else
    result := P2;
end;

function minPoint(P1, P2: TPoint): TPoint;                   //tskurz 12/14/1998
begin
  if P1.y < P2.y then
  begin
    result := P1;
    exit;
  end else
    if P1.y > P2.y then
    begin
      result := P2;
      exit;
    end;
  if P1.x < P2.x then
    result := P1 else
    result := P2;
end;

{ TmwEditList }

function TmwEditList.Add(const S: string): Integer;
begin
  BeginUpdate;
  Result := inherited Add(S);
  if Assigned(FOnAdded) then FOnAdded(Self);
  EndUpdate;
end;

procedure TmwEditList.Delete(Index: Integer);
begin
  BeginUpdate;
  inherited Delete(Index);
  if Assigned(FOnDeleted) then fOnDeleted(Index);
  EndUpdate;
end;

procedure TmwEditList.Insert(Index: Integer; const S: string);
begin
  BeginUpdate;
  inherited Insert(Index, S);
  if Assigned(FOnInserted) then fOnInserted(Index);
  EndUpdate;
end;

procedure TmwEditList.Put(Index: Integer; const S: string);
begin
  BeginUpdate;
  inherited Put(Index, S);
  if Assigned(FOnPutted) then fOnPutted(Index);
  EndUpdate;
end;

procedure TmwEditList.SetTextStr(const Value: string);
begin
  BeginUpdate;
  inherited SetTextStr(Value);
  if Assigned(FOnLoaded) then FOnLoaded(Self);
  EndUpdate;
end;

{ TmwSelectedColor }

constructor TmwSelectedColor.Create(AOwner: TmwCustomEdit);
begin
  inherited Create;
  FOwner := AOwner;
  FBG := clHighLight;
  FFG := clHighLightText;
end;

procedure TmwSelectedColor.SetBG(Value: TColor);
begin
  if (FBG <> Value) then
    begin
      FBG := Value;
      if Assigned(FOwner) then
        FOwner.Invalidate;
    end;
end;

procedure TmwSelectedColor.SetFG(Value: TColor);
begin
  if (FFG <> Value) then
    begin
      FFG := Value;
      if Assigned(FOwner) then
        FOwner.Invalidate;
    end;
end;

{ TmwCustomEdit }

procedure TmwCustomEdit.ComputeCaret(X, Y: Integer);
begin
  CaretX := (X + LeftChar * fCharWidth - FGutterWidth - 2) div fCharWidth;
  CaretY := Y div fTextHeight + TopLine;
end;

procedure TmwCustomEdit.CopyToClipboard;
begin
  if SelAvail then ClipBoard.SetTextBuf(PChar(SelText));
end;

procedure TmwCustomEdit.CutToClipboard;
begin
  if SelAvail then
  begin
    ClipBoard.SetTextBuf(PChar(SelText));
    FUndoList.AddChange(mwcrDelete,fBlockBegin,fBlockEnd,PChar(SelText));
                                                             //tskurz 12/10/1998
    SetSelText('');
  end;
end;

constructor TmwCustomEdit.Create(AOwner: TComponent);
var
  i: Integer;                                                   //jdj 12/10/1998
begin
  fLines := TmwEditList.Create;
  fFontDummy := TFont.Create;
  fUndoList := TUndoList.Create(self);                       //tskurz 12/15/1998
  inherited Create(AOwner);
  {$IFDEF MWE_COMPILER_4_UP}                    //gp 11/09/1998, //mt 12/16/1998
    DoubleBuffered := True;
  {$ENDIF}
  if not (csDesigning in ComponentState) then fLines.Add('');   //jdj 12/08/1998
  TmwEditList(fLines).OnAdded := ListAdded;
  TmwEditList(fLines).OnDeleted := ListDeleted;
  TmwEditList(fLines).OnInserted := ListInserted;
  TmwEditList(fLines).OnLoaded := ListLoaded;
  TmwEditList(fLines).OnPutted := ListPutted;
  fSelectedColor := TmwSelectedColor.Create(Self);              //jdj 12/09/1998
  fBookMarkOpt := TmwBookMarkOpt.Create(Self);                  //jdj 12/12/1998
  fEditTimer := TTimer.Create(Self);
  fEditTimer.Interval := 100;
  feditTimer.OnTimer := TimerFired;
  fCharsInWindow := 10;
  fCharWidth := 8;
  fLinesInWindow := 10;
  fTextHeight := 16;
  fGutterWidth := 30;
  fTextOffset := fGutterWidth + 2;
  ControlStyle := ControlStyle + [csOpaque, csSetCaption];
  Height := 150;
  Width := 200;
  Cursor := crIBeam;
  Color := clWindow;
  fCaretX := 0;
  fCaretY := 0;
  fGutterColor := clBtnFace;
  fFontDummy.Name := 'Courier New';
  fFontDummy.Size := 10;
  Font.Assign(fFontDummy);
  Font.OnChange := FontChanged;
  FontChanged(nil);
  ParentFont := False;
  ParentColor := False;
  TabStop := True;                                              //jdj 12/12/1998
  TStringList(Lines).OnChange := LinesChanged;
  with fGutterRect do begin
    Left := 0;
    Top := 0;
    Right := fGutterWidth;                                       //gp 12/12/1998
    Bottom := 150;
  end;
  fInserting := True;
  fLeftChar := 0;
  fMaxLeftChar := 1024;
  fPaintLock := 0;
  fReadOnly := False;
  fRightEdge := 80;
  fScrollBars := ssBoth;
  fTopLine := 0;
  fBlockBegin := (Point(0, 0));
  fBlockEnd := (Point(0, 0));
  for i := 0 to 9 do                                            //jdj 12/10/1998
    begin
      fBookMarks[i].Glyph := TBitmap.Create;
      fBookMarks[i].Glyph.LoadFromResourceName(HInstance, 'MARK_' + IntToStr(i));
    end;
  fBorderStyle := bsSingle;                                     //bds 12/16/1998
  fHideSelection := false;                                      //bds 12/16/1998
  fScrollPastEOL := true;                                       //bds 12/16/1998
end;

procedure TmwCustomEdit.CreateParams(var Params: TCreateParams);
const
  ScrollBar: array[TScrollStyle] of DWORD = (0, WS_HSCROLL, WS_VSCROLL,
    WS_HSCROLL or WS_VSCROLL);
  BorderStyles: array[TBorderStyle] of DWORD = (0, WS_BORDER);  //bds 12/16/1998
begin
  inherited CreateParams(Params);
  with Params do
  begin
    {begin}                                                     //bds 12/16/1998
    Style := Style or ScrollBar[FScrollBars] or BorderStyles[fBorderStyle]
                   or WS_CLIPCHILDREN;
    if NewStyleControls and Ctl3D and (fBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
    {end}                                                       //bds 12/16/1998
  end;
end;

procedure TmwCustomEdit.DecPaintLock;
begin
  dec(fPaintLock);
  if PaintLock = 0 then Paint;
end;

destructor TmwCustomEdit.Destroy;
var
  i: Integer;                                                   //jdj 12/10/1998
begin
  fFontDummy.Free;
  Lines.Free;
  fSelectedColor.Free;                                          //jdj 12/09/1998
  fBookMarkOpt.Free;                                            //jdj 12/12/1998
  fEditTimer.Free;
  fUndoList.Free;                                            //tskurz 12/10/1998
  for i := 0 to 9 do                                            //jdj 12/10/1998
    fBookMarks[i].Glyph.Free;
  inherited Destroy;
end;

function TmwCustomEdit.GetBlockBegin: TPoint;
begin
  Result := fBlockBegin;
  if fBlockEnd.Y < fBlockBegin.Y then Result := fBlockEnd else
    if fBlockEnd.Y = fBlockBegin.Y then
      if fBlockEnd.X < fBlockBegin.X then Result := fBlockEnd;
end;

function TmwCustomEdit.GetBlockEnd: TPoint;
begin
  Result := fBlockEnd;
  if fBlockEnd.Y < fBlockBegin.Y then Result := fBlockBegin else
    if fBlockEnd.Y = fBlockBegin.Y then
      if fBlockEnd.X < fBlockBegin.X then Result := fBlockBegin;
end;

function TmwCustomEdit.GetCaretX: Integer;
begin
  Result := fCaretX + 1;
end;

function TmwCustomEdit.CaretXPix: Integer;
begin
  Result := fCaretX * fCharWidth + fTextOffset;
end;

function TmwCustomEdit.GetCaretY: Integer;
begin
  Result := fCaretY + 1;
end;

function TmwCustomEdit.CaretYPix: Integer;
begin
  fCaretYPix := fCaretY * fTextHeight - fTopLine * fTextHeight + 1;
  Result := fCaretYPix;
end;

procedure TmwCustomEdit.FontChanged(Sender: TObject);
var
  DC: HDC;
  Save: THandle;
  Metrics: TTextMetric;
begin
  DC := GetDC(0);
  Save := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, Save);
  ReleaseDC(0, DC);
  with Metrics do
  begin
    fCharWidth := tmAveCharWidth;
    fTextHeight := tmHeight + tmExternalLeading;
    if PaintLock = 0 then Invalidate;
  end;
end;

function TmwCustomEdit.GetFont: TFont;
begin
  Result := inherited Font;
end;

function TmwCustomEdit.GetLineCount: Integer;
begin
  Result := Lines.Count;
end;

function TmwCustomEdit.GetLeftChar: Integer;
begin
  Result := fLeftChar + 1;
end;

function TmwCustomEdit.GetLineText: String;
begin
  Result := '';
  if (Lines.Count > 0) and (fCaretY < Lines.Count) then
    Result := Lines[fCaretY];
end;

function TmwCustomEdit.GetSelAvail: Boolean;
begin
  Result := False;
  if (fBlockBegin.X <> fBlockEnd.X) or
    (fBlockBegin.Y <> fBlockEnd.Y) then
    Result := True;
end;

function TmwCustomEdit.GetSelText: String;
var
  BB, BE: TPoint;
  First, Last: Integer;
  Helper: TStringList;
begin
  Result := '';
  if SelAvail then
  begin
    Helper := TStringList.Create;
    BB := BlockBegin;
    BE := BlockEnd;
    if BB.Y = BE.Y then
    begin
      Result := (Copy(Lines[BB.Y - 1], BB.X, BE.X - BB.X));
    end else
    begin
      First := BB.Y - 1;
      Last := BE.Y - 1;
      Helper.Add(Copy(Lines[First], BB.X, Length(Lines[First])));
      inc(First);
      while First < Last do
      begin
        Helper.Add(Lines[First]);
        inc(First);
      end;
      Result := Helper.Text + Copy(Lines[First], 1, BE.X - 1);
    end;
    Helper.Free;
  end;
end;

function TmwCustomEdit.GetText: String;
begin
  Result := Lines.Text;
end;

function TmwCustomEdit.GetTopLine: Integer;
begin
  Result := fTopLine + 1;
end;

procedure TmwCustomEdit.HideCaret;
begin
  if fCaretVisible then
  begin
    if Windows.HideCaret(Handle) then fCaretVisible := False;
  end;
end;

procedure TmwCustomEdit.IncPaintLock;
begin
  inc(fPaintLock);
end;

procedure TmwCustomEdit.KeyDown(var Key: Word; Shift: TShiftState);
var
  moveBkm: boolean;
begin
  {$IFDEF DebugKeyEvents} OutputDebugString('>KeyDown'); try {$ENDIF}
  inherited;
  if Assigned(OnKeyDown) then
    OnKeyDown(Self, Key, Shift);
  Case (ssCtrl in Shift) of
    True:
      begin
        Case Key of
          Ord('C'): CopyToClipboard;
          Ord('V'):
            if not ReadOnly then PasteFromClipboard;
          Ord('X'):
            if not ReadOnly then CutToClipboard;
          Ord('c'): CopyToClipboard;
          Ord('v'):
            if not ReadOnly then PasteFromClipboard;
          Ord('x'):
            if not ReadOnly then CutToClipboard;
          else if ((ssShift in Shift) and (Key in [48..57])
            and fBookMarkOpt.fEnableKeys) then                  //jdj 12/12/1998
          begin
            if fBookMarks[Ord(Key) - 48].IsSet then begin
              moveBkm := (fBookMarks[Ord(Key) - 48].Y <> CaretY);//gp 12/11/1998
              ClearBookMark(Ord(Key) - 48);
              if moveBkm then                                    //gp 12/11/1998
                SetBookMark(Ord(Key) - 48, CaretX, CaretY);
            end
            else SetBookMark(Ord(Key) - 48, CaretX, CaretY);
            Exit;
          end
          else if (Key in [48..57]) 
            and fBookMarkOpt.fEnableKeys then                   //jdj 12/12/1998
            begin                                               //jdj 12/10/1998
              GotoBookMark(Key - 48);
              Exit;
            end;
        end;
      end;
  end;
  case Key of
    VK_UP:     ProcessVK_UP(Shift);
    VK_DOWN:   ProcessVK_DOWN(Shift);
    VK_NEXT:   ProcessVK_NEXT(Shift);
    VK_PRIOR:  ProcessVK_PRIOR(Shift);
    VK_END:    ProcessVK_END(Shift);
    VK_HOME:   ProcessVK_HOME(Shift);
    VK_LEFT:   ProcessVK_LEFT(Shift);
    VK_RIGHT:  ProcessVK_RIGHT(Shift);
    VK_INSERT: ProcessVK_INSERT(Shift);
    VK_DELETE: ProcessVK_DELETE(Shift);
    VK_BACK:   ProcessVK_BACK(Shift); 
    VK_RETURN: ProcessVK_RETURN(Shift);
  end;
  {$IFDEF DebugKeyEvents} finally OutputDebugString('<KeyDown'); end; {$ENDIF}
end;

procedure TmwCustomEdit.Loaded;                                 //jdj 12/12/1998
begin
  inherited Loaded;
  fRightEdgeOffset := fTextOffset + fRightEdge * fCharWidth;
  UpdateScrollBars;
end;

procedure TmwCustomEdit.KeyPress(var Key: Char);
// jdj 12/08/1998
// Changes stop caret from going out of view when character width is less
// than 25
var
  Len: Integer;
  Temp, Helper: string;
  StartOfBlock : TPoint;                                     //tskurz 12/10/1998
  oldScroll: boolean;                                            //gp 12/16/1998
begin
  {$IFDEF DebugKeyEvents} OutputDebugString('>KeyPress'); try {$ENDIF}
  try
    if Assigned(OnKeyPress) then
      OnKeyPress(Self, Key);
    if Key < #32 then Exit;
    if ReadOnly then Exit;
    if SelAvail then
      begin
        FUndoList.AddChange(mwcrDelete,fBlockBegin,fBlockEnd,PChar(SelText));
                                                             //tskurz 12/10/1998
        StartOfBlock := fBlockBegin;                         //tskurz 12/10/1998
        SetSelText(Key);
        FUndoList.AddChange(mwcrInsert,StartOfBlock,fBlockEnd,PChar(SelText));
                                                             //tskurz 12/10/1998
        Exit;
      end;
    Temp := LineText;
    Len := Length(Temp);
    if Len < CaretX then
      begin
        Helper := StringOfChar(' ', CaretX - Len);
        Temp := Temp + Helper;
        Helper := '';
      end;
    case fInserting of
      True:
        begin
          oldScroll := fScrollPastEOL;                           //gp 12/16/1998
          fScrollPastEOL := true;                                //gp 12/16/1998
          try                                                    //gp 12/16/1998
            IncPaintLock;
            StartOfBlock := Point(CaretX,CaretY);            //tskurz 12/10/1998
            Insert(Key, Temp, CaretX);
            CaretX := CaretX + 1;
            LineText := TrimRight(Temp); ;
            FUndoList.AddChange(mwcrInsert,StartOfBlock,Point(CaretX,CaretY),nil);
                                                             //tskurz 12/10/1998
            Temp := '';
            if CaretX >= LeftChar + fCharsInWindow then         //jdj 12/08/1998
              LeftChar := LeftChar + min(25, fCharsInWindow - 1);
            DecPaintLock;
          finally fScrollPastEOL := oldScroll; end;              //gp 12/16/1998
        end;
      False:
        begin
          oldScroll := fScrollPastEOL;                           //gp 12/16/1998
          fScrollPastEOL := true;                                //gp 12/16/1998
          try                                                    //gp 12/16/1998
            IncPaintLock;
            StartOfBlock := Point(CaretX,CaretY);            //tskurz 12/10/1998
            Helper := Temp[CaretX];                          //tskurz 12/10/1998
            Temp[CaretX] := Key;
            CaretX := CaretX + 1;
            LineText := TrimRight(Temp);
            FUndoList.AddChange(mwcrInsert,StartOfBlock,Point(CaretX,CaretY),PChar(Helper));
                                                             //tskurz 12/10/1998
            Temp := '';
            if CaretX >= LeftChar + fCharsInWindow then         //jdj 12/08/1998
              LeftChar := LeftChar + min(25, fCharsInWindow - 1);
            DecPaintLock;
          finally fScrollPastEOL := oldScroll; end;              //gp 12/16/1998
        end;
    end;
  finally
  end;
  {$IFDEF DebugKeyEvents} finally OutputDebugString('<KeyPress'); end; {$ENDIF}
end;

function TmwCustomEdit.LeftSpaces(Line: String): Integer;
var
  Len: Integer;
begin
  Result := 0;
  Len := Length(Line);
  while Result < Len do
  begin
    if Line[Result + 1] > #32 then break;
    inc(Result);
  end;
end;

procedure TmwCustomEdit.LinesChanged(Sender: TObject);
begin
  if HandleAllocated then
  begin
    UpdateScrollBars;
    SetBlockBegin(Point(CaretX, CaretY));
    if PaintLock = 0 then Paint;
  end;
end;

procedure TmwCustomEdit.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);                                                //tk 12/09/1998
var
  bWasSel: boolean;                                             //ked 12/10/1998
  BlStart, BlEnd, Dwn: integer;
begin
  {$IFDEF DebugMouseEvents} OutputDebugString('>MouseDown'); try {$ENDIF}
  {begin}                                                       //ked 12.10.1998
  if SelAvail then begin
    //remember selection state, as it will be cleared later
    bWasSel := true;
    fDragBlockBegin := BlockBegin;
    fDragBlockEnd := BlockEnd;
    fMouseDownX := X;
    fMouseDownY := Y;
  end else
    bWasSel := false;
  {end}                                                         //ked 12/10/1998

  inherited MouseDown(Button, Shift, X, Y);
  MouseCapture := True;
  ComputeCaret(X, Y);
  //if mousedown occured in selected block then begin drag operation
  BlStart := CoorToIndex(fDragBlockBegin);                      //ked 12/10/1998
  BlEnd := CoorToIndex(fDragBlockEnd);
  Dwn := CoorToIndex(Point(CaretX, CaretY));
  fWaitForDragging := false;                                     //gp 12/12/1998
  if bWasSel and (Dwn >= BlStart) and (Dwn < BlEnd)
    then fWaitForDragging := true                                //gp 12/12/1998
    else begin
      if not fDblClicked then begin
        if ssShift in Shift then
          SetBlockEnd(Point(CaretX, CaretY))
        else
          SetBlockBegin(Point(CaretX, CaretY));
      end;
    end;
//  fDblClicked := False;                                        //gp 12/12/1998
  SetFocus;
  {$IFDEF DebugMouseEvents} finally OutputDebugString('<MouseDown'); end; {$ENDIF}
end;

procedure TmwCustomEdit.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  threshold: integer;                                            //gp 12/16/1998
begin
  {$IFDEF DebugMouseMove} OutputDebugString('>MouseMove'); try {$ENDIF}
  inherited MouseMove(Shift, x, y);
  {begin}
  if MouseCapture and fWaitForDragging then begin                //gp 12/12/1998
    {begin}                                                      //gp 12/16/1998
    {$IFDEF MWE_COMPILER_4_UP}
      threshold := Mouse.DragThreshold;
    {$ELSE}
      threshold := 5; // default in D4
    {$ENDIF}
    if (Abs(fMouseDownX - X) >= threshold) or
       (Abs(fMouseDownY - Y) >= threshold) then begin
    {end}                                                        //gp 12/16/1998
      fWaitForDragging := false;
      {$IFDEF DebugDragEvents} OutputDebugString('BeginDrag'); {$ENDIF}
      BeginDrag(false);                                         //ked 12/10/1998
    end;
  end
  {end}                                                          //gp 12/12/1998
  else begin
    Case X > fGutterWidth of
      True: Cursor := crIBeam;
      False: Cursor := crArrow;
    end;
    if (ssLeft in Shift) and MouseCapture then                   //gp 12/12/1998
    begin
      if fInEditRect then
        ComputeCaret(X, Y);
      SetBlockEnd(Point(CaretX, CaretY));
      Application.ProcessMessages;
      while (X < fGutterWidth) and (MouseCapture) and not fInEditRect do
      begin
        if fElapsed then
        begin
          fElapsed := False;
          LeftChar := LeftChar - 1;
          CaretX := LeftChar;
          SetBlockEnd(Point(CaretX, CaretY));
        end;
        Application.ProcessMessages;
      end;

      while (X > ClientWidth) and MouseCapture and not fInEditRect do
      begin
        if fElapsed then
        begin
          fElapsed := False;
          LeftChar := LeftChar + 1;
          CaretX := fLeftChar + fCharsInWindow;
          SetBlockEnd(Point(CaretX, CaretY));
        end;
        Application.ProcessMessages;
      end;

      while (Y < 0) and MouseCapture and not fInEditRect do
      begin
        if Y < -10 then
        begin
          TopLine := TopLine - fLinesInWindow;
          CaretY := TopLine;
          SetBlockEnd(Point(CaretX, CaretY));
          Application.ProcessMessages;
        end else
        begin
          TopLine := TopLine - 1;
          CaretY := TopLine;
          SetBlockEnd(Point(CaretX, CaretY));
          Application.ProcessMessages;
        end;
      end;

      while (Y > ClientHeight) and MouseCapture and not fInEditRect do
      begin
        if Y > (ClientHeight + 10) then
        begin
          TopLine := TopLine + fLinesInWindow;
          CaretY := fTopLine + fLinesInWindow;
          SetBlockEnd(Point(CaretX, CaretY));
          Application.ProcessMessages;
        end else
        begin
          TopLine := TopLine + 1;
          CaretY := fTopLine + fLinesInWindow;
          SetBlockEnd(Point(CaretX, CaretY));
          Application.ProcessMessages;
        end;
      end;
    end;
  end;
  {$IFDEF DebugMouseMove} finally OutputDebugString('<MouseMove'); end; {$ENDIF}
end;

procedure TmwCustomEdit.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  {$IFDEF DebugMouseEvents} OutputDebugString('>MouseUp'); try {$ENDIF}
  inherited MouseUp(Button, Shift, X, Y);
  MouseCapture := False;
  {begin}                                                        //gp 12/12/1998
  if fWaitForDragging and (not fDblClicked) then begin
    ComputeCaret(X,Y);
    SetBlockBegin(Point(CaretX, CaretY));
    SetBlockEnd(Point(CaretX, CaretY));
    fWaitForDragging := false;
  end;
  fDblClicked := False;
  {end}                                                          //gp 12/12/1998
  {$IFDEF DebugMouseEvents} finally OutputDebugString('<MouseUp'); end; {$ENDIF}
end;

procedure TmwCustomEdit.Paint;
begin
  if PaintLock <> 0 then exit;
  PaintControl;
end;

procedure TmwCustomEdit.PaintControl;
var
  TextBM: TBitMap;

  procedure PaintRightEdge;
  begin
    with TextBM.Canvas do
    begin
      Pen.Color := clSilver;
      pen.Width := 1;
      MoveTo(fRightEdgeOffset, 0);
      LineTo(fRightEdgeOffset, ClientHeight);
    end;
  end;

begin
  TextBM := TBitmap.Create;
  try
    TextBM.Canvas.Font.Assign(Font);
    TextBM.Canvas.Brush.Color := Color;
    TextBM.Height := ClientHeight;
    TextBM.Width := ClientWidth;
    TextBM.Canvas.Brush.Color := Color;
    TextBM.Canvas.FillRect(ClientRect);
    if Lines.Count > 0 then
      if Assigned(fHighLighter) then
        PaintHighLighted(TextBM) else
        PaintText(TextBM);
    PaintRightEdge;
    PaintGutter(TextBM);
    if Assigned(fOnPaint) then                                  //jdj 12/12/1998
       fOnPaint(Self,TextBM.Canvas);
    HideCaret;
    Canvas.CopyRect(ClientRect, TextBM.Canvas, ClientRect);
    UpdateCaret;
  finally
    TextBM.Free;
  end;
end;

procedure TmwCustomEdit.PaintGutter(TextBM: TBitMap);
var
  i: Integer;
begin
  with TextBM.Canvas do
    begin
      Brush.Color := fGutterColor;
      FillRect(fGutterRect);
      Pen.Color := clBtnHighlight;
      pen.Width := 1;
      MoveTo(fGutterWidth - 2, 0);
      LineTo(fGutterWidth - 2, ClientHeight);
      Pen.Color := clBtnShadow;
      MoveTo(fGutterWidth - 1, 0);
      LineTo(fGutterWidth - 1, ClientHeight);
      Pen.Color := clWindow;
      pen.Width := 2;
      MoveTo(fGutterWidth + 1, 0);
      LineTo(fGutterWidth + 1, ClientHeight);
      Brush.Color := Color;
//      if (fGutterWidth > fBookMarks[0].Glyph.Width + 1) then
      if (fGutterWidth > fBookMarks[0].Glyph.Width +
        fBookMarkOpt.LeftMargin + 1) then                       //jdj 12/12/1998
        for i := 0 to 9 do                                      //jdj 12/10/1998
          begin
            if fBookMarks[i].IsSet and
              (fBookMarks[i].Y >= TopLine) and
              (fBookMarks[i].Y <= (TopLine + fLinesInWindow)) and
              (fBookMarks[i].Y <= fLines.Count) and             //jdj 12/11/1998
              (fBookMarkOpt.GlyphsVisible) then                 //jdj 12/12/1998
                                                                //jdj 12/12/1998
//            Draw(2, (fBookMarks[i].Y - 1) * fTextHeight - fTopLine * fTextHeight + 1, fBookMarks[i].Glyph)
              Draw(fBookMarkOpt.LeftMargin, (fBookMarks[i].Y - 1) * fTextHeight - fTopLine * fTextHeight + 1, fBookMarks[i].Glyph)
            else if (fBookMarks[i].Y > fLines.Count) then       //jdj 12/11/1998
              fBookMarks[i].IsSet := False;
          end;
    end;
end;

// jdj 12/08/1998:
// Changes selected color to use FSelectedColor
procedure TmwCustomEdit.PaintHighLighted(TextBM: TBitMap);
var
  H, I, J, K, L, M: Integer;
  BB, BE: TPoint;
begin
  fHighLighter.SetCanvas(TextBM.Canvas);
  {begin}                                                       //bds 12/16/1998
  if HideSelection and not Self.Focused then
  begin
    BB.X := -1;
    BB.Y := -1;
    BE.X := -1;
    BE.Y := -1;
  end else begin
  {end}                                                         //bds 12/16/1998
    BB.X := BlockBegin.X - 1;
    BB.Y := BlockBegin.Y - 1;
    BE.X := BlockEnd.X - 1;
    BE.Y := BlockEnd.Y - 1;
  end;                                                          //bds 12/16/1998
  H := 0;
  J := fTopLine;
  K := LeftChar;
  M := LeftChar + fCharsInWindow;
  for I := 0 to fLinesInWindow do
    begin
      with fLineRect do
        begin
          Top := H;
          Bottom := H + fTextHeight;
        end;
      case J < Lines.Count of
        True:
          begin
            if (J < BB.Y) then
              begin
                if J > 0 then
                  fHighLighter.SetRange(Lines.Objects[J])
                else
                  fHighLighter.ReSetRange;
                fHighLighter.SetLine(Lines[J]);
                while not fHighLighter.GetEol do
                  begin
                    TextBM.Canvas.TextOut(fTextOffset + fHighLighter.GetTokenPos * fCharWidth, H, fHighLighter.GetToken);
                    fHighLighter.Next;
                  end;
              end
            else
              begin
                if (J > BB.Y) and (J < BE.Y) then
                  begin
                    TextBM.Canvas.Brush.Color := FSelectedColor.BG;
                                                                //jdj 12/08/1998
                    //clHighLight;
                    TextBM.Canvas.Font.Style := [];
                    TextBM.Canvas.Font.Color := FSelectedColor.FG;
                                                                //jdj 12/08/1998
                    //clHighLightText;
                    TextBM.Canvas.FillRect(fLineRect);
                    TextBM.Canvas.TextOut(fTextOffset, H, Lines[J]);
                  end
                else
                  begin
                    if (J = BB.Y) and (BlockBegin.Y = BlockEnd.Y) then
                      begin
                        if J > 0 then
                          fHighLighter.SetRange(Lines.Objects[J])
                        else
                          fHighLighter.ReSetRange;
                        fHighLighter.SetLine(Lines[J]);
                        while not fHighLighter.GetEol do
                          begin
                            TextBM.Canvas.TextOut(fTextOffset + fHighLighter.GetTokenPos * fCharWidth, H, fHighLighter.GetToken);
                            fHighLighter.Next;
                          end;
                        L := Length(Lines[J]);
                        while (K <= L) and (K < M) do
                          begin
                            if (K <= BE.X) and (K > BB.X) then
                              begin
                                TextBM.Canvas.Brush.Color := FSelectedColor.BG;
                                                                //jdj 12/08/1998
                                 //clHighLight;
                                TextBM.Canvas.Font.Style := [];
                                TextBM.Canvas.Font.Color := FSelectedColor.FG;
                                                                //jdj 12/08/1998
                                //clHighLightText;
                                TextBM.Canvas.TextOut(fTextOffset + (K - 1) * fCharWidth, H, Lines[J][K]);
                              end;
                            Inc(K);
                          end;
                      end
                    else
                      begin
                        if (J = BB.Y) and (BlockBegin.Y <> BlockEnd.Y) then
                          begin
                            TextBM.Canvas.Brush.Color := FSelectedColor.BG;
                                                                //jdj 12/08/1998
                           //clHighLight;
                            TextBM.Canvas.FillRect(fLineRect);
                            if J > 0 then
                              fHighLighter.SetRange(Lines.Objects[J])
                            else
                              fHighLighter.ReSetRange;
                            fHighLighter.SetLine(Lines[J]);
                            while not fHighLighter.GetEol do
                              begin
                                TextBM.Canvas.TextOut(fTextOffset + fHighLighter.GetTokenPos * fCharWidth, H, fHighLighter.GetToken);
                                fHighLighter.Next;
                              end;
                            L := Length(Lines[J]);
                            while (K <= L) and (K < M) do
                              begin
                                if K > BB.X then
                                  begin
                                    TextBM.Canvas.Brush.Color := FSelectedColor.BG;
                                                                //jdj 12/08/1998
                                    //clHighLight;
                                    TextBM.Canvas.Font.Style := [];
                                    TextBM.Canvas.Font.Color := FSelectedColor.FG;
                                                                //jdj 12/08/1998
                                    //clHighLightText;
                                    TextBM.Canvas.TextOut(fTextOffset + (K - 1) * fCharWidth, H, Lines[J][K]);
                                  end;
                                Inc(K);
                              end;
                          end
                        else
                          begin
                            if (J = BE.Y) and (BB.Y <> BE.Y) then
                              begin
                                K := LeftChar;
                                if J > 0 then
                                  fHighLighter.SetRange(Lines.Objects[J])
                                else
                                  fHighLighter.ReSetRange;
                                fHighLighter.SetLine(Lines[J]);
                                while not fHighLighter.GetEol do
                                  begin
                                    TextBM.Canvas.TextOut(fTextOffset + fHighLighter.GetTokenPos * fCharWidth, H, fHighLighter.GetToken);
                                    fHighLighter.Next;
                                  end;
                                L := Length(Lines[J]);
                                while (K <= L) and (K < M) do
                                  begin
                                    if K <= BE.X then
                                      begin
                                        TextBM.Canvas.Brush.Color := FSelectedColor.BG;
                                                                //jdj 12/08/1998
                                        //clHighLight;
                                        TextBM.Canvas.Font.Style := [];
                                        TextBM.Canvas.Font.Color := FSelectedColor.FG;
                                                                //jdj 12/08/1998
                                        //clHighLightText;
                                        TextBM.Canvas.TextOut(fTextOffset + (K - 1) * fCharWidth, H, Lines[J][K]);
                                      end;
                                    Inc(K);
                                  end;
                              end
                            else if (J > BE.Y) then
                              begin
                                if J > 0 then
                                  fHighLighter.SetRange(Lines.Objects[J])
                                else
                                  fHighLighter.ReSetRange;
                                fHighLighter.SetLine(Lines[J]);
                                while not fHighLighter.GetEol do
                                  begin
                                    TextBM.Canvas.TextOut(fTextOffset + fHighLighter.GetTokenPos * fCharWidth, H, fHighLighter.GetToken);
                                    fHighLighter.Next;
                                  end;
                              end;
                          end;
                      end;
                  end;
              end;
          end;
        False:
          begin
            fHighLighter.SetCanvas(nil);
            Exit;
          end;
      end;
      Inc(J);
      Inc(H, fTextHeight);
    end;
  fHighLighter.SetCanvas(nil);
end;

// jdj 12/08/1998:
// Changes selected color to use FSelectedColor
procedure TmwCustomEdit.PaintText(TextBM: TBitMap);
var
  H, I, J, K, L, M: Integer;
  BB, BE: TPoint;
begin
  {begin}                                                       //bds 12/16/1998
  if HideSelection and not Self.Focused then begin
    BB.X := -1;
    BB.Y := -1;
    BE.X := -1;
    BE.Y := -1;
  end else begin
  {end}                                                         //bds 12/16/1998
    BB.X := BlockBegin.X - 1;
    BB.Y := BlockBegin.Y - 1;
    BE.X := BlockEnd.X - 1;
    BE.Y := BlockEnd.Y - 1;
  end;                                                          //bds 12/16/1998
  H := 0;
  J := fTopLine;
  K := LeftChar;
  M := LeftChar + fCharsInWindow;
  for I := 0 to fLinesInWindow do
    begin
      with fLineRect do
        begin
          Top := H;
          Bottom := H + fTextHeight;
        end;
      case J < Lines.Count of
        True:
          begin
            if (J < BB.Y) then
              begin
                TextBM.Canvas.TextOut(fTextOffset, H, Lines[J]);
              end
            else
              begin
                if (J > BB.Y) and (J < BE.Y) then
                  begin
                    TextBM.Canvas.Brush.Color := FSelectedColor.BG;
                                                                //jdj 12/08/1998
                    //clHighLight;
                    TextBM.Canvas.Font.Color := FSelectedColor.FG;
                                                                //jdj 12/08/1998
                    //clHighLightText;
                    TextBM.Canvas.FillRect(fLineRect);
                    TextBM.Canvas.TextOut(fTextOffset, H, Lines[J]);
                  end
                else
                  begin
                    if (J = BB.Y) and (BlockBegin.Y = BlockEnd.Y) then
                      begin
                        L := Length(Lines[J]);
                        while (K <= L) and (K < M) do
                          begin
                            if (K <= BE.X) and (K > BB.X) then
                              begin
                                TextBM.Canvas.Brush.Color := FSelectedColor.BG;
                                                                //jdj 12/08/1998
                                //clHighLight;
                                TextBM.Canvas.Font.Color := FSelectedColor.FG;
                                                                //jdj 12/08/1998
                                //clHighLightText;
                              end
                            else
                              begin
                                TextBM.Canvas.Brush.Color := Color;
                                TextBM.Canvas.Font.Color := Font.Color;
                              end;
                            TextBM.Canvas.TextOut(fTextOffset + (K - 1) * fCharWidth, H, Lines[J][K]);
                            Inc(K);
                          end;
                      end
                    else
                      begin
                        if (J = BB.Y) and (BlockBegin.Y <> BlockEnd.Y) then
                          begin
                            TextBM.Canvas.Brush.Color := FSelectedColor.BG;
                                                                //jdj 12/08/1998
                            //clHighLight;
                            TextBM.Canvas.FillRect(fLineRect);
                            L := Length(Lines[J]);
                            while (K <= L) and (K < M) do
                              begin
                                if K > BB.X then
                                  begin
                                    TextBM.Canvas.Brush.Color := FSelectedColor.BG;
                                                                //jdj 12/08/1998
                                    //clHighLight;
                                    TextBM.Canvas.Font.Color := FSelectedColor.FG;
                                                                //jdj 12/08/1998
                                   //clHighLightText;
                                  end
                                else
                                  begin
                                    TextBM.Canvas.Brush.Color := Color;
                                    TextBM.Canvas.Font.Color := Font.Color;
                                  end;
                                TextBM.Canvas.TextOut(fTextOffset + (K - 1) * fCharWidth, H, Lines[J][K]);
                                Inc(K);
                              end;
                          end
                        else
                          begin
                            if (J = BE.Y) and (BB.Y <> BE.Y) then
                              begin
                                K := LeftChar;
                                L := Length(Lines[J]);
                                while (K <= L) and (K < M) do
                                  begin
                                    if K <= BE.X then
                                      begin
                                        TextBM.Canvas.Brush.Color := FSelectedColor.BG;
                                                                //jdj 12/08/1998
                                        //clHighLight;
                                        TextBM.Canvas.Font.Color := FSelectedColor.FG;
                                                                //jdj 12/08/1998
                                        //clHighLightText;
                                      end
                                    else
                                      begin
                                        TextBM.Canvas.Brush.Color := Color;
                                        TextBM.Canvas.Font.Color := Font.Color;
                                      end;
                                    TextBM.Canvas.TextOut(fTextOffset + (K - 1) * fCharWidth, H, Lines[J][K]);
                                    Inc(K);
                                  end;
                              end
                            else if (J > BE.Y) then
                              begin
                                TextBM.Canvas.Brush.Color := Color;
                                TextBM.Canvas.Font.Color := Font.Color;
                                TextBM.Canvas.TextOut(fTextOffset, H, Lines[J]);
                              end;
                          end;
                      end;
                  end;
              end;
          end;
        False: Exit;
      end;
      Inc(J);
      Inc(H, fTextHeight);
    end;
end;

procedure TmwCustomEdit.PasteFromClipboard;
var
   StartOfBlock: TPoint;                                     //tskurz 12/10/1998
begin
  if Clipboard.HasFormat(CF_TEXT) then
  begin
    StartOfBlock := fBlockBegin;                             //tskurz 12/10/1998
    SelText := Clipboard.AsText;
    FUndoList.AddChange(mwcrInsert,StartOfBlock,fBlockBegin,PChar(SelText));
                                                             //tskurz 12/10/1998
  end;
end;

procedure TmwCustomEdit.ProcessVK_UP(Shift: TShiftState);
begin
  {begin}                                                       //bds 12/16/1998
  if ssCtrl in Shift then
  begin
    TopLine := TopLine - 1;
    if CaretY > TopLine + LinesInWindow - 1 then
      CaretY := TopLine + LinesInWindow - 1;
  end else begin
  {end}                                                         //bds 12/16/1998
    CaretY := CaretY - 1;
    if ssShift in Shift then
    begin
      if not SelAvail then SetBlockBegin(Point(CaretX, CaretY + 1));
      SetBlockEnd(Point(CaretX, CaretY));
    end else SetBlockBegin(Point(CaretX, CaretY));
    if CaretY < TopLine then TopLine := TopLine - 1;
  end;                                                          //bds 12/16/1998
end;

procedure TmwCustomEdit.ProcessVK_DOWN(Shift: TShiftState);
begin
  {begin}                                                       //bds 12/16/1998
  if ssCtrl in Shift then
  begin
    TopLine := TopLine + 1;
    if CaretY < TopLine then
      CaretY := TopLine;
  end else begin
  {end}                                                         //bds 12/16/1998
    CaretY := CaretY + 1;
    if ssShift in Shift then
    begin
      if not SelAvail then SetBlockBegin(Point(CaretX, CaretY - 1));
      SetBlockEnd(Point(CaretX, CaretY));
    end else SetBlockBegin(Point(CaretX, CaretY));
    if CaretY >= TopLine + LinesInWindow then TopLine := TopLine + 1;
  end;                                                          //bds 12/16/1998
end;

procedure TmwCustomEdit.ProcessVK_NEXT(Shift: TShiftState);
begin
  if ssShift in Shift then
    if not SelAvail then SetBlockBegin(Point(CaretX, CaretY));
  CaretY := CaretY + (LinesInWindow - 1);              //jdj 12/12/1998 added -1
  TopLine := TopLine + (LinesInWindow - 1);            //jdj 12/12/1998 added -1
  if ssShift in Shift then
    SetBlockEnd(Point(CaretX, CaretY))
  else                                                          //jdj 12/11/1998
    SetBlockBegin(Point(CaretX, CaretY));
end;

procedure TmwCustomEdit.ProcessVK_PRIOR(Shift: TShiftState);
begin
  if ssShift in Shift then
    if not SelAvail then SetBlockBegin(Point(CaretX, CaretY));
  CaretY := CaretY - (LinesInWindow-1);                //jdj 12/12/1998 added -1
  TopLine := TopLine - (LinesInWindow - 1);            //jdj 12/12/1998 added -1
  if ssShift in Shift then
    SetBlockEnd(Point(CaretX, CaretY))
  else                                                          //jdj 12/11/1998
    SetBlockBegin(Point(CaretX, CaretY));
end;

// jdj 12/09/1998: Scrolls view to right when text extends past right
// bds 12/16/1998: Better positioning of Ctrl-End
procedure TmwCustomEdit.ProcessVK_END(Shift: TShiftState);
var
  CX: Integer;
begin
  if ssShift in Shift then
    if not SelAvail then SetBlockBegin(Point(CaretX, CaretY));
  if not (ssCtrl in Shift) then                                 //jdj 12/11/1998
    begin
      CX := Length(LineText) + 1;                               //jdj 12/08/1998
      if CX >= (LeftChar + fCharsInWindow) then
        begin
          LeftChar := ((CX + 1) - fCharsInWindow);
          CX := LeftChar + fCharsInWindow - 1;
        end
      else if (CX < LeftChar) then
        LeftChar := CX;
      CaretX := CX;
    end
  else
//  if ssCtrl in Shift then                                     //jdj 12/11/1998
    begin
      CaretY := Lines.Count;
      TopLine := Lines.Count - fLinesInWindow + 1;              //bds 12/16/1998
      CaretX := Length(Lines[Lines.Count-1]) + 1;               //bds 12/15/1998
      LeftChar := 1;                                            //jdj 12/11/1998
    end;
  if ssShift in Shift then
    SetBlockEnd(Point(CaretX, CaretY))
  else                                                          //jdj 12/11/1998
    SetBlockBegin(Point(CaretX, CaretY));
end;

procedure TmwCustomEdit.ProcessVK_HOME(Shift: TShiftState);
begin
  if ssShift in Shift then
    if not SelAvail then SetBlockBegin(Point(CaretX, CaretY));
  CaretX := 1;
  if CaretX < LeftChar then LeftChar := CaretX;
  if ssCtrl in Shift then
    begin
      CaretY := 1;
      TopLine := 0;
    end;
  if ssShift in Shift then
    SetBlockEnd(Point(CaretX, CaretY))
  else                                                          //jdj 12/11/1998
    SetBlockBegin(Point(CaretX, CaretY));
end;

// jdj 12/09/1998:
// Smooths caret movement when selecting text
procedure TmwCustomEdit.ProcessVK_LEFT(Shift: TShiftState);
var
  CX: Integer;
begin
  CX := CaretX - 1;                                             //jdj 12/09/1998
//  CaretX := CaretX - 1;
  if ssShift in Shift then
    begin
      if not SelAvail then SetBlockBegin(Point(CX {CaretX} + 1, CaretY));
      SetBlockEnd(Point(CX {CaretX}, CaretY));
    end
  else
    SetBlockBegin(Point(CX {CaretX}, CaretY));
  {begin}                                                        //gp 12/16/1998
  if (not fScrollPastEOL) and (CX < 1) then begin
    if CaretY > 1 then begin
      CaretY := CaretY - 1;
      CaretX := Length(LineText) + 1;
    end;
  end
  {end}                                                          //gp 12/16/1998
  else CaretX := CX;                                            //jdj 12/09/1998
  if CaretX < LeftChar then LeftChar := LeftChar - 1;
end;

// jdj 12/09/1998:
// Smooths caret movement when selecting text
procedure TmwCustomEdit.ProcessVK_RIGHT(Shift: TShiftState);
var
  CX: Integer;
begin
  CX := CaretX + 1;                                             //jdj 12/09/1998
//  CaretX := CaretX + 1;
  if ssShift in Shift then
    begin
      if not SelAvail then SetBlockBegin(Point(CX {CaretX} - 1, CaretY));
      SetBlockEnd(Point(CX {CaretX}, CaretY));
    end
  else
    SetBlockBegin(Point(CX {CaretX}, CaretY));
  {begin}                                                        //gp 12/16/1998
  if (not fScrollPastEOL) and (CX > (Length(LineText) + 1)) then begin
    if CaretY < Lines.Count then begin
      CaretX := 1;
      CaretY := CaretY + 1;
    end;
  end
  {end}                                                          //gp 12/16/1998
  else CaretX := CX;                                            //jdj 12/09/1998
  if CaretX >= LeftChar + fCharsInWindow then LeftChar := LeftChar + 1;
end;

procedure TmwCustomEdit.ProcessVK_INSERT(Shift: TShiftState);
begin
  if ssCtrl in Shift then
  begin
    CopyToClipboard;
  end else
  begin
    if ssShift in Shift then
    begin
      if not ReadOnly then PasteFromClipboard;
    end else
    begin
      Case fInserting of
        True: InsertMode := False;                              //bds 12/16/1998
        False: InsertMode := True;                              //bds 12/16/1998
      end;
    end;
  end;
end;

procedure TmwCustomEdit.ProcessVK_DELETE(Shift: TShiftState);
var
  Len: Integer;
  Temp: string;
  Helper: string;
begin
  if ReadOnly then Exit;
  if SelAvail then
    begin
    {begin}                                                  //tskurz 12/10/1998
    // should have been here since Version 0.18
      if ssShift in Shift then
        begin
          CutToClipboard;
        end
      else
        begin
          FUndoList.AddChange(mwcrDelete, fBlockBegin, fBlockEnd, PChar(SelText));
          SetSelText('');
        end;
    {end}                                                    //tskurz 12/10/1998
      Exit;
    end;
  Temp := LineText;
  Len := Length(Temp);
  if Len >= CaretX then
    begin
      IncPaintLock;
      Helper := Temp[CaretX];                                //tskurz 12/10/1998
      FUndoList.AddChange(mwcrDelete, Point(CaretX, CaretY),   //tskurz 12/10/1998
        Point(CaretX, CaretY), PChar(Helper));
      Delete(Temp, CaretX, 1);
      LineText := TrimRight(Temp);
      Temp := '';
      DecPaintLock;
    end
  else
    begin
      if fCaretY < Lines.Count - 1 {2} then                     //jdj 12/11/1998
        begin
          IncPaintLock;
          Helper := StringOfChar(' ', fCaretX - Len);
          LineText := TrimRight(Temp + Helper + Lines[CaretY]);
          Helper := #13#10;                                  //tskurz 12/10/1998
          FUndoList.AddChange(mwcrDelete, Point(CaretX, CaretY),
            Point(CaretX, CaretY), PChar(Helper));           //tskurz 12/10/1998
          Lines.Delete(CaretY);
          if Lines.Count = 0 then                               //jdj 12/11/1998
            Lines.Add('');
          DecPaintLock;
        end;
    end;
end;

procedure TmwCustomEdit.ProcessVK_BACK(Shift: TShiftState);
var
  Len: Integer;
  SpaceCount1: Integer;
  SpaceCount2: Integer;
  Helper,                                                    //tskurz 12/10/1998
  Temp: string;
  BackCounter: Integer;
begin
  if ReadOnly then Exit;
  // Prevent Alt-Backspace from deleting a character.            //gp 12/12/1998
  if ssAlt in Shift then Exit;
  if SelAvail then
    begin
      FUndoList.AddChange(mwcrDelete, fBlockBegin, fBlockEnd,  //tskurz 12/10/1998
        PChar(SelText));
      SetSelText('');
      Exit;
    end;
  IncPaintLock;
  Temp := LineText;
  Len := Length(Temp);
  SpaceCount2 := 0;
//  if Len > 0 then                                             //jdj 12/11/1998
//  begin
    if Len + 1 >= CaretX then
      begin
        if CaretX > 1 then
          begin
            SpaceCount1 := LeftSpaces(Temp);
            if (Temp[CaretX - 1] <= #32) and (SpaceCount1 = fCaretX) then
              begin
                if SpaceCount1 > 0 then
                  begin
                    BackCounter := fCaretY - 1;
                    if BackCounter < 0 then SpaceCount2 := 0;
                    while BackCounter > 0 do
                      begin
                        SpaceCount2 := LeftSpaces(Lines[BackCounter]);
                        if SpaceCount2 < SpaceCount1 then break;
                        Dec(BackCounter);
                      end;
                  end;
                Helper := Copy(Temp, 1, SpaceCount1 - SpaceCount2);
                                                             //tskurz 12/10/1998
                Delete(Temp, 1, SpaceCount1 - SpaceCount2);
                LineText := TrimRight(Temp);
                Temp := '';
                CaretX := CaretX - (SpaceCount1 - SpaceCount2);
                FUndoList.AddChange(mwcrDelete, Point(CaretX, CaretY),
                  Point(CaretX, CaretY), PChar(Helper));     //tskurz 12/15/1998
                if CaretX < LeftChar then LeftChar := CaretX;
              end
            else
              begin
                CaretX := CaretX - 1;                        //tskurz 12/10/1998
                Helper := Temp[CaretX];                      //tskurz 12/10/1998
                FUndoList.AddChange(mwcrDelete, Point(CaretX, CaretY),
                  Point(CaretX, CaretY), PChar(Helper));     //tskurz 12/10/1998
                Delete(Temp, CaretX, 1);
                LineText := TrimRight(Temp);
                Temp := '';
                if CaretX < LeftChar then LeftChar := CaretX;
              end;
          end
        else
          begin
            if fCaretY > 0 then
              begin
                Lines.Delete(fCaretY);
                CaretY := CaretY - 1;
                CaretX := Length(LineText) + 1;
                Helper := #13#10;                            //tskurz 12/10/1998
                FUndoList.AddChange(mwcrDelete, Point(CaretX, CaretY),
                  Point(CaretX, CaretY), PChar(Helper));     //tskurz 12/10/1998
                LineText := LineText + TrimRight(Temp);
                Temp := '';
                if CaretY < TopLine then TopLine := TopLine - 1;
                if CaretX >= LeftChar + fCharsInWindow then
                  LeftChar := CaretX - fCharsInWindow + 1;
              end;
          end;
      end
    else
      begin
        FUndoList.AddChange(mwcrDelete, Point(CaretX, CaretY),
          Point(CaretX, CaretY), '');                        //tskurz 12/15/1998
        CaretX := CaretX - 1;
        if CaretX < LeftChar then LeftChar := CaretX;
      end;
// I feel a bit strange for taking out this much original code  //jdj 12/11/1998
// but it doesn't seem to be needed and zero length lines
// now appear to be handled properly
(*  end; 
  else
    begin
      if CaretX > 1 then
        begin
          CaretX := 1;
          LeftChar := 1;
        end
      else
        begin
          if (fCaretY > 0) and (fCaretY < Lines.Count - 1) then
            begin
              Lines.Delete(fCaretY);
              Helper := #13#10; {tskurz 12/10/1998}
              FUndoList.AddChange(mwcrDelete, Point(CaretX, CaretY), Point(CaretX, CaretY), PChar(Helper)); {tskurz 12/10/1998}
            end;
          if Len > 0 then
            CaretX := Length(LineText) + 1;
          Temp := '';
        end;
    end;*)
  DecPaintLock;
end;

procedure TmwCustomEdit.ProcessVK_RETURN(Shift: TShiftState);
var
  Len: Integer;
  SpaceCount1: Integer;
  SpaceCount2: Integer;
  Temp: String;
  Temp2: String;
  Helper: String;
  BackCounter: Integer;
begin
  if ReadOnly then exit;
  IncPaintLock;
  SpaceCount2 := 0;
  Temp := LineText;
  Temp2 := LineText;
  Len := Length(Temp);
  if Len > 0 then
  begin
    if Len >= CaretX then
    begin
      if CaretX > 1 then
      begin
        SpaceCount1 := LeftSpaces(Temp);
        if fCaretX <= SpaceCount1 then
        begin
          Lines[fCaretY] := '';
          Lines.Insert(CaretY, Temp);
          FUndoList.AddChange(mwcrLineBreak,Point(CaretX,CaretY),Point(CaretX,fCaretY),PChar(Temp)); {tskurz 12/10/1998}
          CaretY := CaretY + 1;
          if CaretY >= TopLine + LinesInWindow then TopLine := TopLine + 1;
        end else
        begin
          Temp := Copy(LineText, 1, fCaretX);
          LineText := TrimRight(Temp);
          Helper := StringOfChar(' ', SpaceCount1);
          Delete(Temp2, 1, fCaretX);
          FUndoList.AddChange(mwcrLineBreak,Point(CaretX,CaretY),Point(CaretX,CaretY),PChar(Temp2)); {tskurz 12/10/1998}
          Temp2 := Helper + Temp2;
          Lines.Insert(CaretY, Temp2);
          CaretX := SpaceCount1 + 1;
          CaretY := CaretY + 1;
          if CaretY >= TopLine + LinesInWindow then TopLine := TopLine + 1;
        end;
      end else
      begin
        BackCounter := fCaretY - 1;
        while BackCounter > 0 do
        begin
          if Length(Lines[BackCounter]) > 0 then break;
          dec(BackCounter);
        end;
        Lines[fCaretY] := '';
        Lines.Insert(CaretY, Temp);
        FUndoList.AddChange(mwcrLineBreak,Point(CaretX,CaretY),Point(CaretX,CaretY),PChar(Temp2)); {tskurz 12/10/1998}
        CaretY := CaretY + 1;
        if CaretY >= TopLine + LinesInWindow then TopLine := TopLine + 1;
      end;
    end else
    begin
      BackCounter := fCaretY;
      while BackCounter > 0 do
      begin
        SpaceCount2 := LeftSpaces(Lines[BackCounter]);
        if Length(Lines[BackCounter]) > 0 then break;
        dec(BackCounter);
      end;
      FUndoList.AddChange(mwcrLineBreak,Point(CaretX,CaretY),Point(CaretX,CaretY),PChar('')); {tskurz 12/10/1998}
      CaretX := SpaceCount2 + 1;
      Lines.Insert(CaretY, '');
      CaretY := CaretY + 1;
      if CaretY >= TopLine + LinesInWindow then TopLine := TopLine + 1;
    end;
  end else
  begin
    BackCounter := fCaretY;
    while BackCounter >= 0 do
    begin
      SpaceCount2 := LeftSpaces(Lines[BackCounter]);
      if Length(Lines[BackCounter]) > 0 then break;
      dec(BackCounter);
    end;
    FUndoList.AddChange(mwcrLineBreak,Point(CaretX,CaretY),Point(CaretX,CaretY),PChar('')); {tskurz 12/10/1998}
    CaretX := SpaceCount2 + 1;
    Lines.Insert(fCaretY, '');
    CaretY := CaretY + 1;
    if CaretY >= TopLine + LinesInWindow then TopLine := TopLine + 1;
  end;
  DecPaintLock;
end;

procedure TmwCustomEdit.SelectAll;
begin
  SetBlockBegin(Point(1, 1));
  if Lines.Count > 0 then
    SetBlockEnd(Point(Length(Lines[Lines.Count - 1]) + 1, Lines.Count - 1));
end;

procedure TmwCustomEdit.SetBlockBegin(Value: TPoint);
var
  NeedPaint: Boolean;
begin
  if Value.X < 1 then Value.X := 1 else
    if Value.X > fMaxLeftChar then Value.X := fMaxLeftChar;
  if Value.Y < 1 then Value.Y := 1 else
    if Value.Y > Lines.Count then Value.Y := Lines.Count;
  if SelAvail then NeedPaint := True else NeedPaint := False;
  fBlockBegin := Value;
  fBlockEnd := Value;
  if NeedPaint then Paint;
end;

procedure TmwCustomEdit.SetBlockEnd(Value: TPoint);
begin
  if Value.X < 1 then Value.X := 1 else
    if Value.X > fMaxLeftChar then Value.X := fMaxLeftChar;
  if Value.Y < 1 then Value.Y := 1 else
    if Value.Y > Lines.Count then Value.Y := Lines.Count;
  fBlockEnd := Value;
  if PaintLock = 0 then Paint;
end;

// jdj 12/09/1998:
// Smooths caret movement
procedure TmwCustomEdit.SetCaretX(Value: Integer);
var
  Len: integer;
begin
  if Value < 1 then Value := 1;
  {begin}                                                       //bds 12/16/1998
//if Value > (fMaxLeftChar - fCharsInWindow) then Value := fMaxLeftChar - fCharsInWindow;
  if fScrollPastEOL then
  begin
    if Value > (fMaxLeftChar - fCharsInWindow) then
      Value := fMaxLeftChar - fCharsInWindow;
  end else begin
    {begin}                                                      //gp 12/16/1998
    Len := Length(LineText) + 1;
    if Value > Len then
      Value := Length(LineText)+1;
    if Value >= (LeftChar + fCharsInWindow) then
      LeftChar := ((Value + 1) - fCharsInWindow)
    else if Value < LeftChar then LeftChar := Value;
    {end}                                                        //gp 12/16/1998
  end;
  {end}                                                         //bds 12/16/1998
  fCaretX := Value - 1;
  if PaintLock = 0 then                                         //jdj 12/08/1998
    UpdateCaret;
end;

procedure TmwCustomEdit.SetCaretY(Value: Integer);
var
  Len: integer;
begin
  if Value < 1 then Value := 1;
  if Value > Lines.Count then Value := Lines.Count;
  fCaretY := Value - 1;
  {begin}                                                       //bds 12/16/1998
//if PaintLock = 0 then                                         //jdj 12/11/1998
//  UpdateCaret;
  if not FScrollPastEOL then begin
    {begin}                                                      //gp 12/16/1998
    Len := Length(LineText) + 1;
    if CaretX > Len then begin
      CaretX := Len;
      if CaretX < LeftChar then LeftChar := CaretX;              //gp 12/16/1998
      exit;
    end;
    {end}                                                        //gp 12/16/1998
  end;
  if PaintLock = 0 then                                         //jdj 12/11/1998
    UpdateCaret;
  {end}                                                         //bds 12/16/1998
end;

procedure TmwCustomEdit.SetFont(const Value: TFont);
var
  DC: HDC;
  Save: THandle;
  Metrics: TTextMetric;
  AveCW, MaxCW: Integer;
begin
  DC := GetDC(0);
  Save := SelectObject(DC, Value.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, Save);
  ReleaseDC(0, DC);
  with Metrics do
  begin
    AveCW := tmAveCharWidth;
    MaxCW := tmMaxCharWidth;
  end;
  case AveCW = MaxCW of
    True: inherited Font := Value;
    False:
      begin
        with fFontDummy do
        begin
          Color := Value.Color;
          Pitch := fpFixed;
          Size := Value.Size;
          Style := Value.Style;
        end;
        inherited Font := fFontDummy;
      end;
  end;
end;

procedure TmwCustomEdit.SetGutterWidth(Value: Integer);
var
  OldValue: Integer;
begin
  OldValue := GutterWidth;
  if OldValue <> Value then
    begin
      if Value < 0 then
        Value := 0;
      fGutterWidth := Value;
      fGutterRect.Right := Value;
      fCharsInWindow := (ClientWidth - fGutterWidth - 2) div fCharWidth;
      fRightChar := fLeftChar + fCharsInWindow;
      fEditRect := ClientRect;
      fEditRect.Left := fGutterWidth;
      fTextOffset := fGutterWidth + 2;                          //jdj 12/12/1998
      UpdateScrollBars;                                         //jdj 12/12/1998
      if PaintLock = 0 then
        Paint;
    end;
end;

procedure TmwCustomEdit.SetLeftChar(Value: Integer);
var
  OldValue: Integer;
begin
  OldValue := LeftChar;
  if OldValue <> Value then
  begin
    if Value < 1 then Value := 1;
    if Value > fMaxLeftChar - fCharsInWindow then Value := fMaxLeftChar - fCharsInWindow;
    fLeftChar := Value - 1;
    fTextOffset := fGutterWidth + 2 - fLeftChar * fCharWidth;
    fRightEdgeOffset := fTextOffset + fRightEdge * fCharWidth;
    UpdateScrollBars;
    if PaintLock = 0 then Paint;
  end;
end;

procedure TmwCustomEdit.SetLines(Value: TStrings);
begin
  if HandleAllocated then
    Lines.Assign(Value);
end;

procedure TmwCustomEdit.SetLineText(Value: String);
begin
  if (Lines.Count > 0) and
    (fCaretY <= Lines.Count - 1) then Lines[fCaretY] := Value;
end;

// jdj 12/09/1998:
// Stops putting name in control when loading in IDE and Text is empty
procedure TmwCustomEdit.SetName(const Value: TComponentName);
var
  OldName: string;
begin
  OldName := Name;
  inherited SetName(Value);
//  if csDesigning in ComponentState then
  if (csDesigning in ComponentState) and                        //jdj 12/09/1998
  (not (csLoading in ComponentState)) then
    if OldName = TrimRight(Text) then
      Text := Value;
end;

procedure TmwCustomEdit.SetScrollBars(const Value: TScrollStyle);
begin
  if (FScrollBars <> Value) then
  begin
    FScrollBars := Value;
    RecreateWnd;
    UpdateScrollBars;
    if PaintLock = 0 then Paint;
  end;
end;

procedure TmwCustomEdit.SetSelText(Const Value: String);
var
  BB, BE: TPoint;
  TempString: String;
  Temp: TStringList;
  TempY: Integer;

  procedure DeleteSelection;
  var
    I, Count, Start: Integer;
    TempString2: String;
  begin
    IncPaintLock;
    CaretX := BB.X;
    if BB.Y = BE.Y then
    begin
      TempString := Lines[BB.Y - 1];
      Delete(TempString, BB.X, BE.X - BB.X);
      Lines[BB.Y - 1] := TrimRight(TempString);
      TempString := '';
    end else
    begin
      CaretY := BB.Y;
      TempString := Lines[BB.Y - 1];
      Delete(TempString, BB.X, Length(TempString));
      TempString2 := Lines[BE.Y - 1];
      Delete(TempString2, 1, BE.X - 1);
      TempString := TempString + TempString2;
      Start := BB.Y - 1;
      Count := BE.Y - BB.Y;
      for I := 1 to Count do Lines.Delete(Start);
      Lines[Start] := TrimRight(TempString);
    end;
    DecPaintLock;
  end;

  procedure InsertText;
  var
    I, Len: Integer;
    TempString2, Helper: String;
  begin
    if Value = '' then exit;
    IncPaintLock;
    Temp.Text := Value + #13#10;
    if Temp.Count = 1 then
    begin
      TempString := LineText;
      Len := fCaretX - Length(TempString);
      if Len > 0 then
      begin
        Helper := StringOfChar(' ', Len);
        TempString := TempString + Helper;
      end;
      Insert(Temp[0], TempString, CaretX);
      LineText := TempString;
      CaretX := CaretX + Length(Temp[0]);
      TempString := '';
    end else
    begin
      TempString := LineText;
      Len := CaretX - Length(TempString);
      if Len > 0 then
      begin
        Helper := StringOfChar(' ', Len);
        TempString := TempString + Helper;
      end;
      TempString := Copy(TempString, 1, fCaretX) + Temp[0];
      TempString2 := Copy(LineText, CaretX, Length(LineText));
      Lines[fCaretY] := TempString;
      TempString := '';
      TempY := CaretY;
      for I := 1 to Temp.Count - 2 do
      begin
        Lines.Insert(TempY, Temp[I]);
        inc(TempY);
      end;
      Lines.Insert(TempY, Temp[Temp.Count - 1] + TempString2);
      CaretX := Length(Temp[Temp.Count - 1]) + 1;
      inc(TempY);
      CaretY := TempY;
      TempString2 := '';
    end;
    DecPaintLock;
  end;
begin
  IncPaintLock;
  Lines.BeginUpdate;
  Temp := TStringList.Create;
  try
    BB := BlockBegin;
    BE := BlockEnd;
    if SelAvail then
    begin
      DeleteSelection;
      InsertText;
    end else InsertText;
  finally
    Temp.Free;
  end;
  Lines.EndUpdate;
  DecPaintLock;
end;

procedure TmwCustomEdit.SetText(const Value: String);
begin
  Lines.Text := Value;
  if PaintLock = 0 then Invalidate;
end;

procedure TmwCustomEdit.SetTopLine(Value: Integer);
var
  OldValue: Integer;
begin
  OldValue := TopLine;
  if OldValue <> Value then
  begin
    if Value > Lines.Count then Value := Lines.Count;
    if Value < 1 then Value := 1;
    fTopLine := Value - 1;
    UpdateScrollBars;
    if PaintLock = 0 then Paint;
  end;
end;

procedure TmwCustomEdit.ShowCaret;
begin
  if Windows.ShowCaret(Handle) then fCaretVisible := True;
end;

procedure TmwCustomEdit.TimerFired(Sender: TObject);
begin
  fElapsed := not fElapsed;
end;

procedure TmwCustomEdit.UpdateCaret;
var
  CX, CY: Integer;
begin
  if not Focused then Exit;                                     //jdj 12/12/1998
  CX := CaretXPix;
  CY := CaretYPix;
  SetCaretPos(CX, CY);
  ShowCaret;
  if (CX < GutterWidth) or (CX > ClientWidth) or
    (CY < 0) or (CY > ClientHeight - fTextHeight) then
    HideCaret;
end;

procedure TmwCustomEdit.UpdateScrollBars;
var
  ScrollInfo: TScrollInfo;
begin
  if PaintLock <> 0 then exit;
  if fScrollBars in [ssBoth, ssVertical] then
  begin
    ScrollInfo.cbSize := SizeOf(ScrollInfo);
    ScrollInfo.fMask := SIF_ALL;
    ScrollInfo.nMin := 1;
    ScrollInfo.nMax := Lines.Count;
    ScrollInfo.nPage := fLinesInWindow;
    ScrollInfo.nPos := fTopLine;
    ScrollInfo.nTrackPos := 0;
    SetScrollInfo(Handle, SB_VERT, ScrollInfo, True);
    ShowScrollBar(Handle, SB_VERT, True);
  end;
  if fScrollBars in [ssBoth, ssHorizontal] then
  begin
    ScrollInfo.fMask := SIF_ALL;
    ScrollInfo.nMin := 1;
    ScrollInfo.nMax := 1024 - fLinesInWindow;
    ScrollInfo.nPage := fCharsInWindow;
    ScrollInfo.nPos := fLeftChar;
    ScrollInfo.nTrackPos := 0;
    SetScrollInfo(Handle, SB_HORZ, ScrollInfo, True);
    ShowScrollBar(Handle, SB_HORZ, True);
  end;
end;

procedure TmwCustomEdit.WMEraseBkgnd(var Message: TMessage);
begin
  Message.Result := 1;
end;

procedure TmwCustomEdit.WMGetDlgCode(var Msg: TWMGetDlgCode);
begin
  Msg.Result := DLGC_WANTARROWS;
  Msg.Result := Msg.Result or DLGC_WANTCHARS;
end;

procedure TmwCustomEdit.WMHScroll(var Message: TWMScroll);
var
  Delta: Integer;
begin
  Delta := 0;
  case Message.ScrollCode of
    SB_BOTTOM: Delta := fMaxLeftChar; //Scrolls to the lower right.
    SB_ENDSCROLL: Delta := 0; //Ends scroll.
    SB_LINELEFT: Delta := -1; //Scrolls left by one unit.
    SB_LINERIGHT: Delta := 1; //Scrolls right by one unit.
    SB_PAGELEFT: Delta := -fCharsInWindow; //Scrolls left by the width of the window.
    SB_PAGERIGHT: Delta := fCharsInWindow; //Scrolls right by the width of the window.
    SB_THUMBPOSITION: Delta := Message.Pos - LeftChar; //Scrolls to the absolute position. The current position is specified by the nPos parameter.
    SB_THUMBTRACK: Delta := Message.Pos - LeftChar; //Drags scroll box to the specified position. The current position is specified by the nPos parameter.
    SB_TOP: Delta := -fLeftChar; //Scrolls to the upper left.
  end;
  LeftChar := LeftChar + Delta;
end;

procedure TmwCustomEdit.WMKillFocus(var Message: TWMKillFocus);
begin
  inherited;
  HideCaret;
  Windows.DestroyCaret;
  if FHideSelection and SelAvail then Invalidate;               //bds 12/16/1998
end;

procedure TmwCustomEdit.WMNCHitTest(var Message: TWMNCHitTest);
var
  aPoint: TPoint;
begin
  inherited;
  aPoint := ScreenToClient(Point(Message.XPos, Message.YPos));
  if PtInRect(fEditRect, aPoint) then fInEditRect := True else
    fInEditRect := False;
end;

procedure TmwCustomEdit.WMSetFocus(var Message: TWMSetFocus);
begin
  inherited;
//  CreateCaret(Handle, 0, 2, fTextHeight - 2);       //bds 12/16/1998 - removed
//  UpDateCaret;                                      //bds 12/16/1998 - removed
  InitializeCaret;                                              //bds 12/16/1998
  if FHideSelection and SelAvail then                           //bds 12/16/1998
    Invalidate;                                                 //bds 12/16/1998
end;

procedure TmwCustomEdit.WMSize(var Message: TWMSize);
begin
  inherited;
  fCharsInWindow := (ClientWidth - fGutterWidth - 2) div fCharWidth;
  fLinesInWindow := (ClientHeight) div fTextHeight;
  fBottomLine := fTopLine + fLinesInWindow;
  fRightChar := fLeftChar + fCharsInWindow;
  fGutterRect.Bottom := ClientHeight;
  fEditRect := ClientRect;
  fEditRect.Left := fGutterWidth;
  with fLineRect do
  begin
    Left := fGutterWidth;
    Right := ClientWidth;
  end;
end;

procedure TmwCustomEdit.WMVScroll(var Message: TWMScroll);
var
  Delta: Integer;
begin
  Delta := 0;
  case Message.ScrollCode of
    SB_BOTTOM: Delta := Lines.Count - TopLine; //Scrolls to the lower right.
    SB_ENDSCROLL: Delta := 0; //Ends scroll.
    SB_LINEDOWN: Delta := 1; //Scrolls one line down.
    SB_LINEUP: Delta := -1; //Scrolls one line up.
    SB_PAGEDOWN: Delta := fLinesInWindow; //Scrolls one page down.
    SB_PAGEUP: Delta := -fLinesInWindow; //Scrolls one page up.
    SB_THUMBPOSITION: Delta := Message.Pos - TopLine; //Scrolls to the absolute position. The current position is specified by the nPos parameter.
    SB_THUMBTRACK: Delta := Message.Pos - TopLine; //Drags scroll box to the specified position. The current position is specified by the nPos parameter.
    SB_TOP: Delta := -TopLine; // Scrolls to the upper left.
  end;
  TopLine := TopLine + Delta;
end;

procedure TmwCustomEdit.ScanFrom(Index: Integer);
var
  I: Integer;
begin
  if Index >= Lines.Count - 1 then exit;
  I := Index;
  fHighLighter.SetLine(Lines[I]);
  inc(I);
  while not fHighLighter.GetEol do fHighLighter.Next;
  while fHighLighter.GetRange <> fLines.Objects[I] do
  begin
    Lines.Objects[I] := fHighLighter.GetRange;
    fHighLighter.SetLine(Lines[I]);
    while not fHighLighter.GetEol do fHighLighter.Next;
    inc(I);
    if I = Lines.Count then break;
  end;
end;

procedure TmwCustomEdit.ListAdded(Sender: TObject);
var
  LastIndex: Integer;
begin
  if Assigned(fHighLighter) then
  begin
    if Lines.Count > 1 then
    begin
      LastIndex := Lines.Count - 1;
      fHighLighter.SetRange(Lines.Objects[LastIndex - 1]);
      fHighLighter.SetLine(Lines[LastIndex - 1]);
      while not fHighLighter.GetEol do fHighLighter.Next;
      Lines.Objects[LastIndex] := fHighLighter.GetRange;
    end else
    begin
      fHighLighter.ReSetRange;
      Lines.Objects[0] := fHighLighter.GetRange;
    end;
  end;
end;

procedure TmwCustomEdit.ListDeleted(Index: Integer);
var
  OldPaintLock: Integer;                                        //jdj 12/11/1998
begin
  if Assigned(fHighLighter) then
    begin
      if Lines.Count > 1 then
        begin
          if Index > 0 then
            begin
              fHighLighter.SetRange(Lines.Objects[Index - 1]);
              ScanFrom(Index - 1);
            end
          else
            begin
              fHighLighter.ReSetRange;
              Lines.Objects[0] := fHighLighter.GetRange;
              ScanFrom(0);
            end;
        end
      else if Lines.Count = 1 then
        begin
          fHighLighter.ReSetRange;
          Lines.Objects[0] := fHighLighter.GetRange;
        end;
    end;
  {begin}                                                       //jdj 12/11/1998
  OldPaintLock := fPaintLock;
  fPaintLock := 0;
  UpdateScrollBars;
  fPaintLock := OldPaintLock;
  {end}                                                         //jdj 12/11/1998
end;

procedure TmwCustomEdit.ListInserted(Index: Integer);
var
  OldPaintLock: Integer;                                        //jdj 12/11/1998
begin
  if Assigned(fHighLighter) then
    begin
      if Lines.Count > 1 then
        begin
          if Index > 0 then
            begin
              fHighLighter.SetRange(Lines.Objects[Index - 1]);
              ScanFrom(Index - 1);
            end
          else
            begin
              fHighLighter.ReSetRange;
              Lines.Objects[0] := fHighLighter.GetRange;
              ScanFrom(0);
            end;
        end
      else if Lines.Count = 1 then
        begin
          fHighLighter.ReSetRange;
          Lines.Objects[0] := fHighLighter.GetRange;
        end;
    end;
  {begin}                                                       //jdj 12/11/1998
  OldPaintLock := fPaintLock;
  fPaintLock := 0;
  UpdateScrollBars;
  fPaintLock := OldPaintLock;
  {end}                                                         //jdj 12/11/1998
end;

procedure TmwCustomEdit.ListLoaded(Sender: TObject);
var
  i: Integer;
begin
  {begin}                                                       //jdj 12/11/1998
  for i := 0 to 9 do
    fBookMarks[i].IsSet := False;
  CaretX := 1;
  CaretY := 1;
  LeftChar := 1;
  TopLine := 1;
  for i := 0 to fUndoList.FList.Count - 1 do
    fUndoList.RemoveChange(i);
  {end}                                                         //jdj 12/11/1998
  // commented out - I don't know why this is here               //gp 12/13/1998
  //  if Assigned(fHighLighter) then
  //    begin
  //
  //    end;
end;

// mw 12/17/1998 - rewritten
procedure TmwCustomEdit.ListPutted(Index: Integer);
begin
  if Assigned(fHighLighter) then
  begin
    fHighLighter.SetRange(Lines.Objects[Index]);
    ScanFrom(Index);
  end;
end;

procedure TmwCustomEdit.SetWordBlock(Value: TPoint);
var
  Runner: TPoint;
  TempString: String;
begin                                      //tskurz 12/14/1998 - lots of changes
  if Value.X < 1 then Value.X := 1 else
    if Value.X > fMaxLeftChar then Value.X := fMaxLeftChar;
  if Value.Y < 1 then Value.Y := 1 else
    if Value.Y > Lines.Count then Value.Y := Lines.Count;
  TempString := Lines[Value.Y - 1];
  if TempString = '' then exit;

  // Click on right side of text
  if Length(TempString) < Value.X then Value.X := Length(TempString);
  Runner := Value;

//  if not (TempString[Runner.X] in ALPHA+DIGIT) then break;     //mt 12/10/1998
// Add both ALPHA parts because we had to break them up.
  if not (TempString[Runner.X] in ALPHA_UC+ALPHA_LC+DIGIT) then 
  begin
    // no word under cursor and next char right is not start of a word
    if (Runner.X > 1) and (not (TempString[Runner.X] in ALPHA_UC+ALPHA_LC+DIGIT)) then
    begin
      // find end of word on the left side
      while Runner.X > 0 do
      begin
        if (TempString[Runner.X] in ALPHA_UC+ALPHA_LC+DIGIT) then break;
        Dec(Runner.X);
      end;
    end;
    // no word on the left side, so look to the right side
    if not (TempString[Runner.X] in ALPHA_UC+ALPHA_LC+DIGIT) then
    begin
      Runner := Value;
      while Runner.X < fMaxLeftChar do
      begin
        if (TempString[Runner.X] in ALPHA_UC+ALPHA_LC+DIGIT) then break;
        Inc(Runner.X);
      end;
      if Runner.X > fMaxLeftChar then
        exit;
    end;
    Value := Runner;
  end;

  while Runner.X > 0 do
  begin
//  if not (TempString[Runner.X] in ALPHA+DIGIT) then break;     //mt 12/10/1998
// Add both ALPHA parts because we had to break them up.
    if not (TempString[Runner.X] in ALPHA_UC+ALPHA_LC+DIGIT) then break;
    Dec(Runner.X);
  end;
  Inc(Runner.X);
  if Runner.X < 1 then Runner.X := 1;
  fBlockBegin := Runner;

  Runner := Value;
  while Runner.X < fMaxLeftChar do
  begin
//  if not (TempString[Runner.X] in ALPHA+DIGIT) then break;     //mt 12/10/1998
// Add both ALPHA parts because we had to break them up.
    if not (TempString[Runner.X] in ALPHA_UC+ALPHA_LC+DIGIT) then break;
    Inc(Runner.X);
  end;
  if Runner.X > fMaxLeftChar then Runner.X := fMaxLeftChar;
  fBlockEnd := Runner;

  if PaintLock = 0 then Paint;
end;

procedure TmwCustomEdit.DblClick;
begin
  {$IFDEF DebugMouseEvents} OutputDebugString('>DblClick'); try {$ENDIF}
   SetWordBlock(Point(CaretX, CaretY));
   inherited;
   fDblClicked := True;
  {$IFDEF DebugMouseEvents} finally OutputDebugString('<DblClick'); end; {$ENDIF}
end;

function TmwCustomEdit.GetCanUndo: Boolean;
begin
   result := (fUndoList.CanUndo > 0);
end;

procedure TmwCustomEdit.Undo;
var
  ChangeStr : PChar;
  ChangeStartPos,
  ChangeEndPos: TPoint;
  ChangeReason : TChangeReason;
  Temp: String;
begin
  if not CanUndo then exit;
  ChangeReason := FUndoList.GetChange(ChangeStartPos,ChangeEndPos,ChangeStr);
  case ChangeReason of
    mwcrInsert :
      begin
        IncPaintLock;
        SetBlockBegin(ChangeStartPos);
        SetBlockEnd(ChangeEndPos);
        DecPaintLock;
        SelText := ChangeStr;
        StrDispose(ChangeStr);
        CaretX := ChangeStartPos.x;
        CaretY := ChangeStartPos.y;
      end;
    mwcrDragDropInsert :                                     //tskurz 12/16/1998
      begin
        IncPaintLock;
        SetBlockBegin(ChangeStartPos);
        SetBlockEnd(ChangeEndPos);
        DecPaintLock;
        SelText := ChangeStr;
        StrDispose(ChangeStr);
        CaretX := ChangeStartPos.x;
        CaretY := ChangeStartPos.y;
        if FUndoList.GetChangeReason = mwcrDragDropDelete then Undo;
      end;
    mwcrDelete,
    mwcrDragDropDelete :  //tskurz 12/16/1998
      begin
        IncPaintLock;
        // If there's no selection, we have to set
        // the Caret's position manualy.
        CaretX := ChangeStartPos.x;
        CaretY := ChangeStartPos.y;
        SetBlockBegin(ChangeStartPos);
        DecPaintLock;
        SelText := ChangeStr;
        StrDispose(ChangeStr);
      end;
    mwcrLineBreak :
      begin
        IncPaintLock;
        // If there's no selection, we have to set
        // the Caret's position manualy.
        CaretY := ChangeStartPos.y;
        if CaretY > 0 then
        begin
          Temp := Lines.Strings[fCaretY] + ChangeStr;
          Lines.Delete(ChangeEndPos.y);
        end;
        CaretY := ChangeStartPos.y;
        CaretX := ChangeStartPos.x;
        LineText := TrimRight(Temp);
        //CaretX := ChangeStartPos.x;
        DecPaintLock;
        StrDispose(ChangeStr);
      end;
  else
  end;
end;
{end}                                                        //tskurz 12/10/1998

{begin}                                                         //jdj 12/10/1998
procedure TmwCustomEdit.ClearBookMark(BookMark: Integer);
begin
  if (BookMark in [0..9]) then
    begin
      fBookMarks[BookMark].IsSet := False;
      Invalidate;
    end;
end;

procedure TmwCustomEdit.GotoBookMark(BookMark: Integer);
begin
  if (BookMark in [0..9]) and
    fBookMarks[BookMark].IsSet and
    (fBookMarks[BookMark].Y <= fLines.Count) then
    begin
      if (fBookMarks[BookMark].Y < TopLine) or
        (fBookMarks[BookMark].Y >= (TopLine + fLinesInWindow)) then
        TopLine := fBookMarks[BookMark].Y - (fLinesInWindow div 2);
      CaretX := fBookMarks[BookMark].X;
      if (fBookMarks[BookMark].X < LeftChar) or
        (fBookMarks[BookMark].X >= LeftChar + fCharsInWindow) then
        LeftChar := fBookMarks[BookMark].X - (fCharsInWindow div 2);
      CaretY := fBookMarks[BookMark].Y;
    end;
end;

procedure TmwCustomEdit.SetBookMark(BookMark: Integer; X: Integer; Y: Integer);
var
  i: Integer;
begin
  if (BookMark in [0..9]) and (Y <= fLines.Count) then
    begin
      for i := 0 to 9 do
        if (fBookMarks[i].Y = Y) then
          fBookMarks[i].IsSet := False;
      fBookMarks[BookMark].IsSet := True;
      fBookMarks[BookMark].X := X;
      fBookMarks[BookMark].Y := Y;
      Invalidate;
    end;
end;
{end}                                                           //jdj 12/10/1998

procedure TmwCustomEdit.WndProc(var msg: TMessage);              //gp 12/12/1998
// Prevent Alt-Backspace from beeping and wire it to the Undo at the same time.
begin
  if (msg.Msg = WM_SYSCHAR) AND (msg.wParam = VK_BACK) and
     ((msg.lParam AND $20000000) = $20000000) then begin
    Undo;
    msg.Msg := 0;
  end
  else inherited;
end;

{begin}                                                         //ked 12/10/1998
function TmwCustomEdit.CoorToIndex(CPos: TPoint): integer;
begin
  result := CharsInWindow * CPos.Y + CPos.X;
end;

function TmwCustomEdit.IndexToCoor(ind: integer): TPoint;       
begin
  result.y := ind div CharsInWindow;
  result.x := ind - result.y * CharsInWindow;
end;

procedure TmwCustomEdit.DragOver(Source: TObject; X, Y: Integer;
                             State: TDragState; var Accept: Boolean);
begin
  {$IFDEF DebugDragEvents} try OutputDebugString(PChar(Format('>DragOver %p %d %d %d',[pointer(Source),x,y,Ord(State)]))); {$ENDIF}
  inherited;
  if Accept or(Source is TmwCustomEdit) then begin
    Accept := True;
    if ((GetKeyState(VK_CONTROL) shr 31)=0) then begin
      if DragCursor<>crDrag then
        DragCursor := crDrag;
    end else //Ctrl is pressed => change cursor to indicate copy instead of move
      if DragCursor<>crMultiDrag then
        DragCursor := crMultiDrag;
    if State=dsDragLeave then //restore prev caret position
      ComputeCaret(FMouseDownX, FMouseDownY)
    else //position caret under the mouse cursor
      ComputeCaret(X, Y);
  end;
  {$IFDEF DebugDragEvents} finally OutputDebugString('<DragOver'); end; {$ENDIF}
end;

procedure TmwCustomEdit.DragDrop(Source: TObject; X, Y: Integer);
var
  CurInd, DragBegInd, DragEndInd: Integer;
  sWK: string;
  dropLine, oldLines, correction: integer;                       //gp 12/17/1998
  oldScroll: boolean;                                            //gp 12/16/1998
begin
  {$IFDEF DebugDragEvents} try OutputDebugString(PChar(Format('>DragDrop %p %d %d',[pointer(Source),x,y]))); {$ENDIF}
  IncPaintLock;
  try
    inherited;
    if (Source=self) and (SelAvail) then begin
      ComputeCaret(X,Y);
      sWk := SelText; //text to be moved/copied
      CurInd := CoorToIndex(Point(CaretX, CaretY));
      DragBegInd := CoorToIndex(fDragBlockBegin);
      DragEndInd := CoorToIndex(fDragBlockEnd);
      //copy or move text if drop occured out of selection
      if (CurInd<DragBegInd) or (CurInd>DragEndInd) then begin
        if (GetKeyState(VK_CONTROL) shr 31)=0 then begin
          //if Ctrl is not down then text will be moved
          if (DragBegInd<CurInd) and (FDragBlockEnd.Y = CaretY) then
            //take into account deleted text if dropping occured on the same line
            CurInd := CurInd - (DragEndInd - DragBegInd);
          FUndoList.AddChange(mwcrDragDropDelete, fBlockBegin,
            fBlockEnd, PChar(SelText));                      //tskurz 12/16/1998
          oldLines := Lines.Count;                               //gp 12/16/1998
          dropLine := CaretY;                                    //gp 12/17/1998
          SelText := '';
        {begin}                                                  //gp 12/16/1998
          if dropLine > CaretY
            then correction := oldLines-Lines.Count
            else correction := 0;
        end
        else correction := 0;
        {end}                                                    //gp 12/16/1998
        //move
        BlockBegin := IndexToCoor(CurInd);
        {begin}                                                  //gp 12/16/1998
        BlockBegin := Point(BlockBegin.X,BlockBegin.Y-correction);           
        CurInd := CoorToIndex(BlockBegin);                       
        oldScroll := fScrollPastEOL;                             
        fScrollPastEOL := true;
        try                                                      
        {end}                                                    //gp 12/16/1998
          CaretX := BlockBegin.x;
          CaretY := BlockBegin.Y;
          BlockEnd := BlockBegin;
          SelText := sWk;
        finally fScrollPastEOL := oldScroll; end;                //gp 12/16/1998
        FUndoList.AddChange(mwcrDragDropInsert,IndexToCoor(CurInd),
          BlockEnd,PChar(SelText));                          //tskurz 12/16/1998
      end;
    end;
  finally
    DecPaintLock;
  end;
  {$IFDEF DebugDragEvents} finally OutputDebugString('<DragDrop'); end; {$ENDIF}
end;
{end}                                                           //ked 12/10/1998

procedure TmwCustomEdit.SetGutterColor(Value: TColor);          //jdj 12/12/1998
begin
  if fGutterColor <> Value then
    begin
      FGutterColor := Value;
      Invalidate;
    end;
end;

procedure TmwCustomEdit.SetRightEdge(Value: Integer);           //jdj 12/12/1998
begin
  if (fRightEdge <> Value) then
    begin
      fRightEdge := Value;
      fRightEdgeOffset := fTextOffset + fRightEdge * fCharWidth;
      Invalidate;
    end;
end;

{ TmwBookMarkOpt }
{begin}                                                         //jdj 12/12/1998
constructor TmwBookMarkOpt.Create(AOwner: TmwCustomEdit);
begin
  inherited Create;
  FOwner := AOwner;
  fEnableKeys := True;
  fGlyphsVisible := True;
  fLeftMargin := 2;
end;

procedure TmwBookMarkOpt.SetGlyphsVisible(Value: Boolean);
begin
  if (fGlyphsVisible <> Value) then
    begin
      fGlyphsVisible := Value;
      FOwner.Invalidate;
    end;
end;

procedure TmwBookMarkOpt.SetLeftMargin(Value: Integer); 
begin
  if (fLeftMargin <> Value) then
    begin
      fLeftMargin := Value;
      FOwner.Invalidate;
    end;
end;
{end}                                                           //jdj 12/12/1998

function TmwCustomEdit.GetMaxUndo: Integer;                  //tskurz 12/14/1998
begin
   result := fUndoList.MaxUndo;
end;

procedure TmwCustomEdit.SetMaxUndo(const Value: Integer);    //tskurz 12/14/1998
begin
  if Value > -1 then fUndoList.MaxUndo := Value;
end;

procedure TmwCustomEdit.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = fHighLighter) then
    fHighLighter := nil;
end;

procedure TmwCustomEdit.SetHighlighter(const Value: TmwCustomHighLighter);
begin
  fHighLighter := Value;
  if Value <> nil then Value.FreeNotification(Self);
end;

{begin}                                                         //bds 12/16/1998
procedure TmwCustomEdit.SetBorderStyle(Value: TBorderStyle);
begin
  if fBorderStyle <> Value then
  begin
    fBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TmwCustomEdit.SetHideSelection(const Value: boolean);
begin
  if fHideSelection <> Value then
  begin
    FHideSelection := Value;
    Invalidate;
  end;
end;

procedure TmwCustomEdit.SetInsertMode(const Value: boolean);
begin
  if fInserting <> Value then
  begin
    fInserting := Value;
    if not (csDesigning in ComponentState) then
      // Reset the caret.
      InitializeCaret;
  end;
end;

procedure TmwCustomEdit.InitializeCaret;
begin
  // CreateCaret automatically destroys the previous one, so we don't have to
  // worry about cleaning up the old one here with DestroyCaret.
  // Ideally, we will have properties that control what these two carets look like.
  if InsertMode then
    CreateCaret(Handle, 0, 2, fTextHeight - 2)
  else
    CreateCaret(Handle, 0, fCharWidth, fTextHeight - 2);
  UpdateCaret;
end;

procedure TmwCustomEdit.SetScrollPastEOL(const Value: boolean);
begin
  if FScrollPastEOL <> Value then
  begin
    FScrollPastEOL := Value;
    if not FScrollPastEOL then
      CaretX := CaretX;  // Reset pos in case it's past EOL currently.
  end;
end;
{end}                                                           //bds 12/16/1998

{ TUndoList }
{begin}                                                      //tskurz 12/10/1998
procedure TUndoList.AddChange(ChangeReason: TChangeReason; ChangeStartPos,
              ChangeEndPos: TPoint; ChangeStr: PChar);
var
  chngptr : TChangePtr;
begin
   if FList.Count >= FMaxUndo then RemoveChange(0);
   try
      GetMem(chngptr,SizeOf(TChange));
      try
        chngptr^.ChangeStr := StrNew(ChangeStr);
      except
        FreeMem(chngptr);
        exit;
      end;
   except
     exit;
   end;
   chngptr^.ChangeReason := ChangeReason;
   {begin}                                                   //tskurz 12/14/1998
   //chngptr^.ChangeStartPos := ChangeStartPos;
   //chngptr^.ChangeEndPos := ChangeEndPos;
   chngptr^.ChangeStartPos := minPoint(ChangeStartPos,ChangeEndPos);
   chngptr^.ChangeEndPos := maxPoint(ChangeStartPos,ChangeEndPos);
   {end}                                                     //tskurz 12/14/1998
   FList.Add(chngptr);
   if Assigned(FOwner.OnChange) then FOwner.OnChange(FOwner);//tskurz 12/15/1998
end;

constructor TUndoList.Create(AOwner: TmwCustomEdit);         //tskurz 12/15/1998
begin
   inherited Create;
   FOwner := AOwner;                                         //tskurz 12/15/1998
   FList := TList.Create;
   FMaxUndo := 10;
end;

destructor TUndoList.Destroy;
var
   i : Integer;
begin
   for i := 0 to FList.Count - 1 do
     RemoveChange(i);
   FList.Free;
   inherited Destroy;
end;

function TUndoList.GetChange(var ChangeStartPos, ChangeEndPos: TPoint;
  var ChangeStr: PChar): TChangeReason;
begin
   if FList.Count = 0 then
   begin
     result := mwcrNone;
     exit;
   end;
   ChangeStartPos := TChangePtr(FList.Items[FList.Count-1])^.ChangeStartPos;
   ChangeEndPos := TChangePtr(FList.Items[FList.Count-1])^.ChangeEndPos;
   ChangeStr := StrNew(TChangePtr(FList.Items[FList.Count-1])^.ChangeStr);
   result := TChangePtr(FList.Items[FList.Count-1])^.ChangeReason;
   RemoveChange(FList.Count-1);
end;

function TUndoList.GetCanUndo : Integer;
begin
  FCanUndo := FList.Count;
  result := FCanUndo;
end;

procedure TUndoList.SetMaxUndo(const Value: Integer);
var
  i : Integer;
begin
  if Value < FList.Count - 1 then
     for i := 0 to FList.Count - Value - 1 do
        RemoveChange(0);
  FMaxUndo := Value;
end;

procedure TUndoList.RemoveChange(index: Integer);
var
  chngptr : TChangePtr;
begin
  if (index > -1) and (index < FList.Count) then
  begin
    chngptr := FList.Items[Index];
    try
      StrDispose(chngptr^.ChangeStr);
      try
        FreeMem(chngptr);
      except
        exit;
      end;
    except
      exit;
    end;
    Flist.Delete(index);
  end;
end;

function TUndoList.GetChangeReason: TChangeReason;           //tskurz 12/16/1998
begin
   if FList.Count = 0 then
     result := mwcrNone
   else
     result := TChangePtr(FList.Items[FList.Count-1])^.ChangeReason;
end;

end.

